@@ -39,7 +39,7 @@
 #include <csignal>
 #include <cstring>
 
-#ifdef _MSC_VER
+#ifdef _WIN32
 #include "../win32/wingetopt.h"
 #elif _AIX
 #include <unistd.h>
@@ -116,11 +116,11 @@ static void metadata_print (const std::string &topic,
   }
 }
 
-static bool run = true;
+static volatile sig_atomic_t run = 1;
 static bool exit_eof = false;
 
 static void sigterm (int sig) {
-  run = false;
+  run = 0;
 }
 
 
@@ -159,7 +159,7 @@ class ExampleEventCb : public RdKafka::EventCb {
       case RdKafka::Event::EVENT_ERROR:
         if (event.fatal()) {
           std::cerr << "FATAL ";
-          run = false;
+          run = 0;
         }
         std::cerr << "ERROR (" << RdKafka::err2str(event.err()) << "): " <<
             event.str() << std::endl;
@@ -237,20 +237,20 @@ void msg_consume(RdKafka::Message* message, void* opaque) {
     case RdKafka::ERR__PARTITION_EOF:
       /* Last message */
       if (exit_eof) {
-        run = false;
+        run = 0;
       }
       break;
 
     case RdKafka::ERR__UNKNOWN_TOPIC:
     case RdKafka::ERR__UNKNOWN_PARTITION:
       std::cerr << "Consume failed: " << message->errstr() << std::endl;
-      run = false;
+      run = 0;
       break;
 
     default:
       /* Errors */
       std::cerr << "Consume failed: " << message->errstr() << std::endl;
-      run = false;
+      run = 0;
   }
 }
 
@@ -407,7 +407,7 @@ int main (int argc, char **argv) {
             "                  random (default), hash\n"
             "  -b <brokers>    Broker address (localhost:9092)\n"
             "  -z <codec>      Enable compression:\n"
-            "                  none|gzip|snappy\n"
+            "                  none|gzip|snappy|lz4|zstd\n"
             "  -o <offset>     Start offset (consumer)\n"
             "  -e              Exit consumer when last message\n"
             "                  in partition has been received.\n"
@@ -551,7 +551,7 @@ int main (int argc, char **argv) {
 
       producer->poll(0);
     }
-    run = true;
+    run = 1;
 
     while (run && producer->outq_len() > 0) {
       std::cerr << "Waiting for " << producer->outq_len() << std::endl;
