@@ -26,23 +26,27 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 
-#pragma once
+#ifndef _RDKAFKA_METADATA_H_
+#define _RDKAFKA_METADATA_H_
 
 #include "rdavl.h"
 
-struct rd_kafka_metadata *
+rd_kafka_resp_err_t
 rd_kafka_parse_Metadata (rd_kafka_broker_t *rkb,
-                         rd_kafka_buf_t *request, rd_kafka_buf_t *rkbuf);
+                         rd_kafka_buf_t *request, rd_kafka_buf_t *rkbuf,
+                         struct rd_kafka_metadata **mdp);
 
 struct rd_kafka_metadata *
 rd_kafka_metadata_copy (const struct rd_kafka_metadata *md, size_t size);
 
 size_t
 rd_kafka_metadata_topic_match (rd_kafka_t *rk, rd_list_t *tinfos,
-                               const rd_kafka_topic_partition_list_t *match);
+                               const rd_kafka_topic_partition_list_t *match,
+                               rd_kafka_topic_partition_list_t *errored);
 size_t
 rd_kafka_metadata_topic_filter (rd_kafka_t *rk, rd_list_t *tinfos,
-                                const rd_kafka_topic_partition_list_t *match);
+                                const rd_kafka_topic_partition_list_t *match,
+                                rd_kafka_topic_partition_list_t *errored);
 
 void rd_kafka_metadata_log (rd_kafka_t *rk, const char *fac,
                             const struct rd_kafka_metadata *md);
@@ -51,11 +55,17 @@ void rd_kafka_metadata_log (rd_kafka_t *rk, const char *fac,
 
 rd_kafka_resp_err_t
 rd_kafka_metadata_refresh_topics (rd_kafka_t *rk, rd_kafka_broker_t *rkb,
-                                  const rd_list_t *topics, int force,
+                                  const rd_list_t *topics, rd_bool_t force,
+                                  rd_bool_t allow_auto_create,
+                                  rd_bool_t cgrp_update,
                                   const char *reason);
 rd_kafka_resp_err_t
 rd_kafka_metadata_refresh_known_topics (rd_kafka_t *rk, rd_kafka_broker_t *rkb,
-                                        int force, const char *reason);
+                                        rd_bool_t force, const char *reason);
+rd_kafka_resp_err_t
+rd_kafka_metadata_refresh_consumer_topics (rd_kafka_t *rk,
+                                           rd_kafka_broker_t *rkb,
+                                           const char *reason);
 rd_kafka_resp_err_t
 rd_kafka_metadata_refresh_brokers (rd_kafka_t *rk, rd_kafka_broker_t *rkb,
                                    const char *reason);
@@ -66,9 +76,21 @@ rd_kafka_metadata_refresh_all (rd_kafka_t *rk, rd_kafka_broker_t *rkb,
 rd_kafka_resp_err_t
 rd_kafka_metadata_request (rd_kafka_t *rk, rd_kafka_broker_t *rkb,
                            const rd_list_t *topics,
+                           rd_bool_t allow_auto_create_topics,
+                           rd_bool_t cgrp_update,
                            const char *reason, rd_kafka_op_t *rko);
 
 
+
+int rd_kafka_metadata_partition_id_cmp (const void *_a,
+                                        const void *_b);
+
+rd_kafka_metadata_t *
+rd_kafka_metadata_new_topic_mock (const rd_kafka_metadata_topic_t *topics,
+                                  size_t topic_cnt);
+rd_kafka_metadata_t *rd_kafka_metadata_new_topic_mockv (size_t topic_cnt, ...);
+
+
 /**
  * @{
  *
@@ -84,8 +106,15 @@ struct rd_kafka_metadata_cache_entry {
         /* rkmce_partitions memory points here. */
 };
 
-#define RD_KAFKA_METADATA_CACHE_VALID(rkmce) \
-        ((rkmce)->rkmce_mtopic.err != RD_KAFKA_RESP_ERR__WAIT_CACHE)
+
+#define RD_KAFKA_METADATA_CACHE_ERR_IS_TEMPORARY(ERR)                   \
+        ((ERR) == RD_KAFKA_RESP_ERR__WAIT_CACHE ||                      \
+         (ERR) == RD_KAFKA_RESP_ERR__NOENT)
+
+#define RD_KAFKA_METADATA_CACHE_VALID(rkmce)                            \
+        !RD_KAFKA_METADATA_CACHE_ERR_IS_TEMPORARY((rkmce)->rkmce_mtopic.err)
+
+
 
 struct rd_kafka_metadata_cache {
         rd_avl_t         rkmc_avl;
@@ -93,6 +122,9 @@ struct rd_kafka_metadata_cache {
         rd_kafka_timer_t rkmc_expiry_tmr;
         int              rkmc_cnt;
 
+        /* Protected by rk_lock */
+        rd_list_t        rkmc_observers; /**< (rd_kafka_enq_once_t*) */
+
         /* Protected by full_lock: */
         mtx_t            rkmc_full_lock;
         int              rkmc_full_topics_sent; /* Full MetadataRequest for
@@ -114,17 +146,21 @@ struct rd_kafka_metadata_cache {
 void rd_kafka_metadata_cache_expiry_start (rd_kafka_t *rk);
 void
 rd_kafka_metadata_cache_topic_update (rd_kafka_t *rk,
-                                      const rd_kafka_metadata_topic_t *mdt);
+                                      const rd_kafka_metadata_topic_t *mdt,
+                                      rd_bool_t propagate);
 void rd_kafka_metadata_cache_update (rd_kafka_t *rk,
                                      const rd_kafka_metadata_t *md,
                                      int abs_update);
+void rd_kafka_metadata_cache_propagate_changes (rd_kafka_t *rk);
 struct rd_kafka_metadata_cache_entry *
 rd_kafka_metadata_cache_find (rd_kafka_t *rk, const char *topic, int valid);
 void rd_kafka_metadata_cache_purge_hints (rd_kafka_t *rk,
                                           const rd_list_t *topics);
 int rd_kafka_metadata_cache_hint (rd_kafka_t *rk,
                                   const rd_list_t *topics, rd_list_t *dst,
-                                  int replace);
+                                  rd_kafka_resp_err_t err,
+                                  rd_bool_t replace);
+
 int rd_kafka_metadata_cache_hint_rktparlist (
         rd_kafka_t *rk,
         const rd_kafka_topic_partition_list_t *rktparlist,
@@ -143,9 +179,6 @@ int rd_kafka_metadata_cache_topic_partition_get (
 int rd_kafka_metadata_cache_topics_count_exists (rd_kafka_t *rk,
                                                  const rd_list_t *topics,
                                                  int *metadata_agep);
-int rd_kafka_metadata_cache_topics_filter_hinted (rd_kafka_t *rk,
-                                                  rd_list_t *dst,
-                                                  const rd_list_t *src);
 
 void rd_kafka_metadata_fast_leader_query (rd_kafka_t *rk);
 
@@ -154,4 +187,12 @@ void rd_kafka_metadata_cache_destroy (rd_kafka_t *rk);
 int  rd_kafka_metadata_cache_wait_change (rd_kafka_t *rk, int timeout_ms);
 void rd_kafka_metadata_cache_dump (FILE *fp, rd_kafka_t *rk);
 
+int rd_kafka_metadata_cache_topics_to_list (rd_kafka_t *rk,
+                                             rd_list_t *topics);
+
+void
+rd_kafka_metadata_cache_wait_state_change_async (rd_kafka_t *rk,
+                                                 rd_kafka_enq_once_t *eonce);
+
 /**@}*/
+#endif /* _RDKAFKA_METADATA_H_ */
