@@ -55,6 +55,11 @@
  * for topics, but for any topics not currently in the cache a new
  * entry will be added with a flag (RD_KAFKA_METADATA_CACHE_VALID(rkmce))
  * indicating that the entry is waiting to be populated by the MetadataResponse.
+ * Two special error codes are used for this purpose:
+ *   RD_KAFKA_RESP_ERR__NOENT - to indicate that a topic needs to be queried,
+ *   RD_KAFKA_RESP_ERR__WAIT_CACHE - to indicate that a topic is being queried
+ *                                   and there is no need to re-query it prior
+ *                                   to the current query finishing.
  *
  * The cache is locked in its entirety with rd_kafka_wr/rdlock() by the caller
  * and the returned cache entry must only be accessed during the duration
@@ -62,7 +67,6 @@
  *
  */
 
-static void rd_kafka_metadata_cache_propagate_changes (rd_kafka_t *rk);
 
 
 /**
@@ -300,10 +304,18 @@ void rd_kafka_metadata_cache_expiry_start (rd_kafka_t *rk) {
 /**
  * @brief Update the metadata cache for a single topic
  *        with the provided metadata.
- *        If the topic has an error the existing entry is removed
- *        and no new entry is added, which avoids the topic to be
- *        suppressed in upcoming metadata requests because being in the cache.
- *        In other words: we want to re-query errored topics.
+ *
+ * If the topic has a temporary error the existing entry is removed
+ * and no new entry is added, which avoids the topic to be
+ * suppressed in upcoming metadata requests because being in the cache.
+ * In other words: we want to re-query errored topics.
+ * If the broker reports ERR_UNKNOWN_TOPIC_OR_PART we add a negative cache
+ * entry with an low expiry time, this is so that client code (cgrp) knows
+ * the topic has been queried but did not exist, otherwise it would wait
+ * forever for the unknown topic to surface.
+ *
+ * For permanent errors (authorization failures), we keep
+ * the entry cached for metadata.max.age.ms.
  *
  * @remark The cache expiry timer will not be updated/started,
  *         call rd_kafka_metadata_cache_expiry_start() instead.
@@ -312,18 +324,26 @@ void rd_kafka_metadata_cache_expiry_start (rd_kafka_t *rk) {
  */
 void
 rd_kafka_metadata_cache_topic_update (rd_kafka_t *rk,
-                                      const rd_kafka_metadata_topic_t *mdt) {
+                                      const rd_kafka_metadata_topic_t *mdt,
+                                      rd_bool_t propagate) {
         rd_ts_t now = rd_clock();
         rd_ts_t ts_expires = now + (rk->rk_conf.metadata_max_age_ms * 1000);
         int changed = 1;
 
-        if (!mdt->err)
+        /* Cache unknown topics for a short while (100ms) to allow the cgrp
+         * logic to find negative cache hits. */
+        if (mdt->err == RD_KAFKA_RESP_ERR_UNKNOWN_TOPIC_OR_PART)
+                ts_expires = RD_MIN(ts_expires, now + (100 * 1000));
+
+        if (!mdt->err ||
+            mdt->err == RD_KAFKA_RESP_ERR_TOPIC_AUTHORIZATION_FAILED ||
+            mdt->err == RD_KAFKA_RESP_ERR_UNKNOWN_TOPIC_OR_PART)
                 rd_kafka_metadata_cache_insert(rk, mdt, now, ts_expires);
         else
                 changed = rd_kafka_metadata_cache_delete_by_name(rk,
                                                                  mdt->topic);
 
-        if (changed)
+        if (changed && propagate)
                 rd_kafka_metadata_cache_propagate_changes(rk);
 }
 
@@ -364,7 +384,7 @@ void rd_kafka_metadata_cache_update (rd_kafka_t *rk,
                                      rd_kafka_metadata_cache_evict_tmr_cb,
                                      rk);
 
-        if (md->topic_cnt > 0)
+        if (md->topic_cnt > 0 || abs_update)
                 rd_kafka_metadata_cache_propagate_changes(rk);
 }
 
@@ -416,8 +436,10 @@ void rd_kafka_metadata_cache_purge_hints (rd_kafka_t *rk,
  *
  * @param dst rd_list_t(char *topicname): if not NULL: populated with
  *        topics that were added as hints to cache, e.q., topics to query.
- * @param topics rd_list_t(char *topicname)
- * @param replace int: replace existing valid entries
+ * @param dst rd_list_t(char *topicname)
+ * @param err is the error to set on hint cache entries,
+ *            typically ERR__WAIT_CACHE.
+ * @param replace replace existing valid entries
  *
  * @returns the number of topic hints inserted.
  *
@@ -425,7 +447,8 @@ void rd_kafka_metadata_cache_purge_hints (rd_kafka_t *rk,
  */
 int rd_kafka_metadata_cache_hint (rd_kafka_t *rk,
                                   const rd_list_t *topics, rd_list_t *dst,
-                                  int replace) {
+                                  rd_kafka_resp_err_t err,
+                                  rd_bool_t replace) {
         const char *topic;
         rd_ts_t now = rd_clock();
         rd_ts_t ts_expires = now + (rk->rk_conf.socket_timeout_ms * 1000);
@@ -435,16 +458,19 @@ int rd_kafka_metadata_cache_hint (rd_kafka_t *rk,
         RD_LIST_FOREACH(topic, topics, i) {
                 rd_kafka_metadata_topic_t mtopic = {
                         .topic = (char *)topic,
-                        .err = RD_KAFKA_RESP_ERR__WAIT_CACHE
+                        .err = err
                 };
-                const struct rd_kafka_metadata_cache_entry *rkmce;
+                /*const*/ struct rd_kafka_metadata_cache_entry *rkmce;
 
                 /* !replace: Dont overwrite valid entries */
                 if (!replace &&
                     (rkmce =
                      rd_kafka_metadata_cache_find(rk, topic, 0/*any*/))) {
-                        if (RD_KAFKA_METADATA_CACHE_VALID(rkmce) || dst)
+                        if (RD_KAFKA_METADATA_CACHE_VALID(rkmce) ||
+                            (dst && rkmce->rkmce_mtopic.err !=
+                             RD_KAFKA_RESP_ERR__NOENT))
                                 continue;
+                        rkmce->rkmce_mtopic.err = err;
                         /* FALLTHRU */
                 }
 
@@ -480,7 +506,9 @@ int rd_kafka_metadata_cache_hint_rktparlist (
         rd_list_init(&topics, rktparlist->cnt, rd_free);
         rd_kafka_topic_partition_list_get_topic_names(rktparlist, &topics,
                                                       0/*dont include regex*/);
-        r = rd_kafka_metadata_cache_hint(rk, &topics, dst, replace);
+        r = rd_kafka_metadata_cache_hint(rk, &topics, dst,
+                                         RD_KAFKA_RESP_ERR__WAIT_CACHE,
+                                         replace);
         rd_list_destroy(&topics);
         return r;
 }
@@ -507,7 +535,8 @@ void rd_kafka_metadata_cache_init (rd_kafka_t *rk) {
         mtx_init(&rk->rk_metadata_cache.rkmc_full_lock, mtx_plain);
         mtx_init(&rk->rk_metadata_cache.rkmc_cnd_lock, mtx_plain);
         cnd_init(&rk->rk_metadata_cache.rkmc_cnd);
-
+        rd_list_init(&rk->rk_metadata_cache.rkmc_observers, 8,
+                     rd_kafka_enq_once_trigger_destroy);
 }
 
 /**
@@ -516,6 +545,7 @@ void rd_kafka_metadata_cache_init (rd_kafka_t *rk) {
  * @locks rd_kafka_wrlock()
  */
 void rd_kafka_metadata_cache_destroy (rd_kafka_t *rk) {
+        rd_list_destroy(&rk->rk_metadata_cache.rkmc_observers);
         rd_kafka_timer_stop(&rk->rk_timers,
                             &rk->rk_metadata_cache.rkmc_query_tmr, 1/*lock*/);
         rd_kafka_metadata_cache_purge(rk);
@@ -526,6 +556,20 @@ void rd_kafka_metadata_cache_destroy (rd_kafka_t *rk) {
 }
 
 
+
+/**
+ * @brief Add eonce to list of async cache observers.
+ *
+ * @locks_required rd_kafka_wrlock()
+ */
+void
+rd_kafka_metadata_cache_wait_state_change_async (rd_kafka_t *rk,
+                                                 rd_kafka_enq_once_t *eonce) {
+        rd_kafka_enq_once_add_source(eonce, "wait metadata cache change");
+        rd_list_add(&rk->rk_metadata_cache.rkmc_observers, eonce);
+}
+
+
 /**
  * @brief Wait for cache update, or timeout.
  *
@@ -553,16 +597,38 @@ int rd_kafka_metadata_cache_wait_change (rd_kafka_t *rk, int timeout_ms) {
         return r == thrd_success;
 }
 
+
+/**
+ * @brief eonce trigger callback for rd_list_apply() call in
+ *        rd_kafka_metadata_cache_propagate_changes()
+ */
+static int
+rd_kafka_metadata_cache_propagate_changes_trigger_eonce (void *elem,
+                                                         void *opaque) {
+        rd_kafka_enq_once_t *eonce = elem;
+        rd_kafka_enq_once_trigger(eonce, RD_KAFKA_RESP_ERR_NO_ERROR,
+                                  "wait metadata cache change");
+        return 0; /* remove eonce from list */
+}
+
+
 /**
  * @brief Propagate that the cache changed (but not what changed) to
- *        any cnd listeners.
- * @locks none
+ *        any cnd listeners and eonce observers.
+ * @locks_required rd_kafka_wrlock(rk)
+ * @locks_acquired rkmc_cnd_lock
  * @locality any
  */
-static void rd_kafka_metadata_cache_propagate_changes (rd_kafka_t *rk) {
+void rd_kafka_metadata_cache_propagate_changes (rd_kafka_t *rk) {
         mtx_lock(&rk->rk_metadata_cache.rkmc_cnd_lock);
         cnd_broadcast(&rk->rk_metadata_cache.rkmc_cnd);
         mtx_unlock(&rk->rk_metadata_cache.rkmc_cnd_lock);
+
+        /* Trigger observers */
+        rd_list_apply(&rk->rk_metadata_cache.rkmc_observers,
+                      rd_kafka_metadata_cache_propagate_changes_trigger_eonce,
+                      NULL);
+
 }
 
 /**
@@ -588,6 +654,9 @@ rd_kafka_metadata_cache_topic_get (rd_kafka_t *rk, const char *topic,
 /**
  * @brief Looks up the shared metadata for a partition along with its topic.
  *
+ * Cache entries with errors (such as auth errors) will not be returned unless
+ * \p valid is set to false.
+ *
  * @param mtopicp: pointer to topic metadata
  * @param mpartp: pointer to partition metadata
  * @param valid: only return valid entries (no hints)
@@ -615,6 +684,9 @@ int rd_kafka_metadata_cache_topic_partition_get (
 
         *mtopicp = mtopic;
 
+        if (mtopic->err)
+                return -1;
+
         /* Partitions array may be sparse so use bsearch lookup. */
         mpart = bsearch(&skel, mtopic->partitions,
                         mtopic->partition_cnt,
@@ -668,39 +740,35 @@ int rd_kafka_metadata_cache_topics_count_exists (rd_kafka_t *rk,
 
 
 /**
- * @brief Copies any topics in \p src to \p dst that have a valid cache
- *        entry, or not in the cache at all.
+ * @brief Add all topics in the metadata cache to \p topics, avoid duplicates.
  *
- *        In other words; hinted non-valid topics will not copied to \p dst.
+ * Element type is (char *topic_name).
  *
- * @returns the number of topics copied
+ * @returns the number of elements added to \p topics
  *
- * @locks rd_kafka_*lock()
+ * @locks_required rd_kafka_*lock()
  */
-int rd_kafka_metadata_cache_topics_filter_hinted (rd_kafka_t *rk,
-                                                  rd_list_t *dst,
-                                                  const rd_list_t *src) {
-        const char *topic;
-        int i;
-        int cnt = 0;
+int rd_kafka_metadata_cache_topics_to_list (rd_kafka_t *rk,
+                                             rd_list_t *topics) {
+        const struct rd_kafka_metadata_cache_entry *rkmce;
+        int precnt = rd_list_cnt(topics);
 
-
-        RD_LIST_FOREACH(topic, src, i) {
-                const struct rd_kafka_metadata_cache_entry *rkmce;
-
-                rkmce = rd_kafka_metadata_cache_find(rk, topic, 0/*any sort*/);
-                if (rkmce && !RD_KAFKA_METADATA_CACHE_VALID(rkmce))
+        TAILQ_FOREACH(rkmce, &rk->rk_metadata_cache.rkmc_expiry, rkmce_link) {
+                /* Ignore topics that have up to date metadata info */
+                if (RD_KAFKA_METADATA_CACHE_VALID(rkmce))
                         continue;
 
-                rd_list_add(dst, rd_strdup(topic));
-                cnt++;
+                if (rd_list_find(topics, rkmce->rkmce_mtopic.topic,
+                                 rd_list_cmp_str))
+                        continue;
+
+                rd_list_add(topics, rd_strdup(rkmce->rkmce_mtopic.topic));
         }
 
-        return cnt;
+        return rd_list_cnt(topics) - precnt;
 }
 
 
-
 /**
  * @brief Dump cache to \p fp
  *
