@@ -298,11 +298,16 @@ struct rd_kafka_toppar_s { /* rd_kafka_toppar_t */
 
         rd_kafka_resp_err_t rktp_last_error;     /**< Last Fetch error.
                                                   *   Used for suppressing
-                                                  *   reoccuring errors. */
+                                                  *   reoccuring errors.
+                                                  * @locality broker thread */
 
         rd_kafka_cgrp_t   *rktp_cgrp;            /* Belongs to this cgrp */
 
-        int                rktp_assigned;   /* Partition in cgrp assignment */
+        rd_bool_t          rktp_started;  /**< Fetcher is instructured to
+                                           *   start.
+                                           *   This is used by cgrp to keep
+                                           *   track of whether the toppar has
+                                           *   been started or not. */
 
         rd_kafka_replyq_t  rktp_replyq; /* Current replyq+version
 					 * for propagating
@@ -389,19 +394,27 @@ struct rd_kafka_toppar_s { /* rd_kafka_toppar_t */
 /**
  * @brief Increase refcount and return rktp object.
  */
+#define rd_kafka_toppar_keep(RKTP)                              \
+        rd_kafka_toppar_keep0(__FUNCTION__,__LINE__,RKTP)
+
+#define rd_kafka_toppar_keep_fl(FUNC,LINE,RKTP)                 \
+        rd_kafka_toppar_keep0(FUNC,LINE,RKTP)
+
 static RD_UNUSED RD_INLINE
-rd_kafka_toppar_t *rd_kafka_toppar_keep (rd_kafka_toppar_t *rktp) {
-        rd_refcnt_add(&rktp->rktp_refcnt);
+rd_kafka_toppar_t *rd_kafka_toppar_keep0 (const char *func, int line,
+                                          rd_kafka_toppar_t *rktp) {
+        rd_refcnt_add_fl(func, line, &rktp->rktp_refcnt);
         return rktp;
 }
 
 void rd_kafka_toppar_destroy_final (rd_kafka_toppar_t *rktp);
 
-static RD_UNUSED RD_INLINE
-void rd_kafka_toppar_destroy (rd_kafka_toppar_t *rktp) {
-        if (unlikely(rd_refcnt_sub(&rktp->rktp_refcnt) == 0))
-                rd_kafka_toppar_destroy_final(rktp);
-}
+#define rd_kafka_toppar_destroy(RKTP) do {                              \
+                rd_kafka_toppar_t *_RKTP = (RKTP);                      \
+                if (unlikely(rd_refcnt_sub(&_RKTP->rktp_refcnt) == 0))  \
+                        rd_kafka_toppar_destroy_final(_RKTP);           \
+        } while (0)
+
 
 
 
@@ -531,6 +544,9 @@ void rd_kafka_toppar_leader_unavailable (rd_kafka_toppar_t *rktp,
 void rd_kafka_toppar_pause (rd_kafka_toppar_t *rktp, int flag);
 void rd_kafka_toppar_resume (rd_kafka_toppar_t *rktp, int flag);
 
+rd_kafka_resp_err_t
+rd_kafka_toppar_op_pause_resume (rd_kafka_toppar_t *rktp, int pause, int flag,
+                                 rd_kafka_replyq_t replyq);
 rd_kafka_resp_err_t
 rd_kafka_toppars_pause_resume (rd_kafka_t *rk,
                                rd_bool_t pause, rd_async_t async, int flag,
@@ -539,11 +555,24 @@ rd_kafka_toppars_pause_resume (rd_kafka_t *rk,
 
 rd_kafka_topic_partition_t *rd_kafka_topic_partition_new (const char *topic,
 							  int32_t partition);
+void rd_kafka_topic_partition_destroy_free (void *ptr);
+rd_kafka_topic_partition_t *
+rd_kafka_topic_partition_copy (const rd_kafka_topic_partition_t *src);
+void *rd_kafka_topic_partition_copy_void (const void *src);
+void rd_kafka_topic_partition_destroy_free (void *ptr);
 rd_kafka_topic_partition_t *
 rd_kafka_topic_partition_new_from_rktp (rd_kafka_toppar_t *rktp);
 
+void rd_kafka_topic_partition_list_init (
+        rd_kafka_topic_partition_list_t *rktparlist, int size);
+void rd_kafka_topic_partition_list_destroy_free (void *ptr);
+
+void rd_kafka_topic_partition_list_clear (
+        rd_kafka_topic_partition_list_t *rktparlist);
+
 rd_kafka_topic_partition_t *
-rd_kafka_topic_partition_list_add0 (rd_kafka_topic_partition_list_t *rktparlist,
+rd_kafka_topic_partition_list_add0 (const char *func, int line,
+                                    rd_kafka_topic_partition_list_t *rktparlist,
                                     const char *topic, int32_t partition,
 				    rd_kafka_toppar_t *_private);
 
@@ -552,17 +581,59 @@ rd_kafka_topic_partition_list_upsert (
         rd_kafka_topic_partition_list_t *rktparlist,
         const char *topic, int32_t partition);
 
+void rd_kafka_topic_partition_list_add_copy (
+        rd_kafka_topic_partition_list_t *rktparlist,
+        const rd_kafka_topic_partition_t *rktpar);
+
+
+void rd_kafka_topic_partition_list_add_list (
+        rd_kafka_topic_partition_list_t *dst,
+        const rd_kafka_topic_partition_list_t *src);
+
+/**
+ * Traverse rd_kafka_topic_partition_list_t.
+ *
+ * @warning \p TPLIST modifications are not allowed.
+ */
+#define RD_KAFKA_TPLIST_FOREACH(RKTPAR,TPLIST)                          \
+        for (RKTPAR = &(TPLIST)->elems[0] ;                             \
+             (RKTPAR) < &(TPLIST)->elems[(TPLIST)->cnt] ;               \
+             RKTPAR++)
+
+/**
+ * Traverse rd_kafka_topic_partition_list_t.
+ *
+ * @warning \p TPLIST modifications are not allowed, but removal of the
+ *          current \p RKTPAR element is allowed.
+ */
+#define RD_KAFKA_TPLIST_FOREACH_REVERSE(RKTPAR,TPLIST)                  \
+        for (RKTPAR = &(TPLIST)->elems[(TPLIST)->cnt-1] ;               \
+             (RKTPAR) >= &(TPLIST)->elems[0] ;                          \
+             RKTPAR--)
+
 int rd_kafka_topic_partition_match (rd_kafka_t *rk,
 				    const rd_kafka_group_member_t *rkgm,
 				    const rd_kafka_topic_partition_t *rktpar,
 				    const char *topic, int *matched_by_regex);
 
 
+int rd_kafka_topic_partition_cmp (const void *_a, const void *_b);
+unsigned int rd_kafka_topic_partition_hash (const void *a);
+
+int
+rd_kafka_topic_partition_list_find_idx (
+        const rd_kafka_topic_partition_list_t *rktparlist,
+        const char *topic, int32_t partition);
+rd_kafka_topic_partition_t *
+rd_kafka_topic_partition_list_find_topic (
+        const rd_kafka_topic_partition_list_t *rktparlist, const char *topic);
+
 void rd_kafka_topic_partition_list_sort_by_topic (
         rd_kafka_topic_partition_list_t *rktparlist);
 
 void
-rd_kafka_topic_partition_list_reset_offsets (rd_kafka_topic_partition_list_t *rktparlist,
+rd_kafka_topic_partition_list_reset_offsets (rd_kafka_topic_partition_list_t
+                                             *rktparlist,
 					     int64_t offset);
 
 int rd_kafka_topic_partition_list_set_offsets (
@@ -573,9 +644,20 @@ int rd_kafka_topic_partition_list_set_offsets (
 int rd_kafka_topic_partition_list_count_abs_offsets (
 	const rd_kafka_topic_partition_list_t *rktparlist);
 
+int
+rd_kafka_topic_partition_list_cmp (const void *_a, const void *_b,
+                                   int (*cmp) (const void *, const void *));
+
+rd_kafka_toppar_t *
+rd_kafka_topic_partition_ensure_toppar (rd_kafka_t *rk,
+                                        rd_kafka_topic_partition_t *rktpar,
+                                        rd_bool_t create_on_miss);
+
 rd_kafka_toppar_t *
 rd_kafka_topic_partition_get_toppar (rd_kafka_t *rk,
-                                     rd_kafka_topic_partition_t *rktpar);
+                                     rd_kafka_topic_partition_t *rktpar,
+                                     rd_bool_t create_on_miss)
+        RD_WARN_UNUSED_RESULT;
 
 void
 rd_kafka_topic_partition_list_update_toppars (rd_kafka_t *rk,
@@ -583,11 +665,15 @@ rd_kafka_topic_partition_list_update_toppars (rd_kafka_t *rk,
                                               *rktparlist,
                                               rd_bool_t create_on_miss);
 
-int
-rd_kafka_topic_partition_list_get_leaders (
+
+void
+rd_kafka_topic_partition_list_query_leaders_async (
         rd_kafka_t *rk,
-        rd_kafka_topic_partition_list_t *rktparlist,
-        rd_list_t *leaders, rd_list_t *query_topics);
+        const rd_kafka_topic_partition_list_t *rktparlist,
+        int timeout_ms,
+        rd_kafka_replyq_t replyq,
+        rd_kafka_op_cb_t *cb,
+        void *opaque);
 
 rd_kafka_resp_err_t
 rd_kafka_topic_partition_list_query_leaders (
@@ -644,6 +730,11 @@ rd_kafka_topic_partition_list_sum (
         size_t (*cb) (const rd_kafka_topic_partition_t *rktpar, void *opaque),
         void *opaque);
 
+rd_bool_t
+rd_kafka_topic_partition_list_has_duplicates (
+        rd_kafka_topic_partition_list_t *rktparlist,
+        rd_bool_t ignore_partition);
+
 void rd_kafka_topic_partition_list_set_err (
         rd_kafka_topic_partition_list_t *rktparlist,
         rd_kafka_resp_err_t err);
@@ -654,6 +745,9 @@ rd_kafka_resp_err_t rd_kafka_topic_partition_list_get_err (
 int rd_kafka_topic_partition_list_regex_cnt (
         const rd_kafka_topic_partition_list_t *rktparlist);
 
+void *
+rd_kafka_topic_partition_list_copy_opaque (const void *src, void *opaque);
+
 /**
  * @brief Toppar + Op version tuple used for mapping Fetched partitions
  *        back to their fetch versions.
@@ -731,6 +825,8 @@ rd_kafka_partition_leader_destroy (struct rd_kafka_partition_leader *leader) {
         rd_free(leader);
 }
 
+void rd_kafka_partition_leader_destroy_free (void *ptr);
+
 static RD_UNUSED struct rd_kafka_partition_leader *
 rd_kafka_partition_leader_new (rd_kafka_broker_t *rkb) {
         struct rd_kafka_partition_leader *leader = rd_malloc(sizeof(*leader));
