@@ -36,12 +36,19 @@
 #include "rdkafka_transport_int.h"
 #include "rdkafka_cert.h"
 
-#ifdef _MSC_VER
+#ifdef _WIN32
 #pragma comment (lib, "crypt32.lib")
 #endif
 
 #include <openssl/x509.h>
 
+#include <ctype.h>
+
+#if !_WIN32
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <unistd.h>
+#endif
 
 
 #if WITH_VALGRIND
@@ -76,7 +83,7 @@ void rd_kafka_transport_ssl_close (rd_kafka_transport_t *rktrans) {
 static RD_INLINE void
 rd_kafka_transport_ssl_clear_error (rd_kafka_transport_t *rktrans) {
         ERR_clear_error();
-#ifdef _MSC_VER
+#ifdef _WIN32
         WSASetLastError(0);
 #else
         rd_set_errno(0);
@@ -125,7 +132,11 @@ static char *rd_kafka_ssl_error (rd_kafka_t *rk, rd_kafka_broker_t *rkb,
         int line, flags;
         int cnt = 0;
 
-        while ((l = ERR_get_error_line_data(&file, &line, &data, &flags)) != 0) {
+        if (!rk)
+                rk = rkb->rkb_rk;
+
+        while ((l = ERR_get_error_line_data(&file, &line,
+                                            &data, &flags)) != 0) {
                 char buf[256];
 
                 if (cnt++ > 0) {
@@ -138,13 +149,25 @@ static char *rd_kafka_ssl_error (rd_kafka_t *rk, rd_kafka_broker_t *rkb,
 
                 ERR_error_string_n(l, buf, sizeof(buf));
 
-                rd_snprintf(errstr, errstr_size, "%s:%d: %s: %s",
-                            file, line, buf, (flags & ERR_TXT_STRING) ? data : "");
+                if (!(flags & ERR_TXT_STRING) || !data || !*data)
+                        data = NULL;
 
+                /* Include openssl file:line if debugging is enabled */
+                if (rk->rk_conf.log_level >= LOG_DEBUG)
+                        rd_snprintf(errstr, errstr_size, "%s:%d: %s%s%s",
+                                    file, line, buf,
+                                    data ? ": " : "",
+                                    data ? data : "");
+                else
+                        rd_snprintf(errstr, errstr_size, "%s%s%s",
+                                    buf,
+                                    data ? ": " : "",
+                                    data ? data : "");
         }
 
         if (cnt == 0)
-                rd_snprintf(errstr, errstr_size, "No error");
+                rd_snprintf(errstr, errstr_size,
+                            "No further error information available");
 
         return errstr;
 }
@@ -177,11 +200,11 @@ rd_kafka_transport_ssl_io_update (rd_kafka_transport_t *rktrans, int ret,
 
         case SSL_ERROR_SYSCALL:
                 serr2 = ERR_peek_error();
-                if (!serr2 && !rd_socket_errno)
-                        rd_snprintf(errstr, errstr_size, "Disconnected");
-                else if (serr2)
+                if (serr2)
                         rd_kafka_ssl_error(NULL, rktrans->rktrans_rkb,
                                            errstr, errstr_size);
+                else if (!rd_socket_errno || rd_socket_errno == ECONNRESET)
+                        rd_snprintf(errstr, errstr_size, "Disconnected");
                 else
                         rd_snprintf(errstr, errstr_size,
                                     "SSL transport error: %s",
@@ -587,11 +610,36 @@ int rd_kafka_transport_ssl_handshake (rd_kafka_transport_t *rktrans) {
         } else if (rd_kafka_transport_ssl_io_update(rktrans, r,
                                                     errstr,
                                                     sizeof(errstr)) == -1) {
+                const char *extra = "";
+
+                if (strstr(errstr, "unexpected message"))
+                        extra = ": client SSL authentication might be "
+                                "required (see ssl.key.location and "
+                                "ssl.certificate.location and consult the "
+                                "broker logs for more information)";
+                else if (strstr(errstr, "tls_process_server_certificate:"
+                                "certificate verify failed") ||
+                         strstr(errstr, "get_server_certificate:"
+                                "certificate verify failed"))
+                        extra = ": broker certificate could not be verified, "
+                                "verify that ssl.ca.location is correctly "
+                                "configured or root CA certificates are "
+                                "installed"
+#ifdef __APPLE__
+                                " (brew install openssl)"
+#elif defined(_WIN32)
+                                " (add broker's CA certificate to the Windows "
+                                "Root certificate store)"
+#else
+                                " (install ca-certificates package)"
+#endif
+                                ;
+                else if (!strcmp(errstr, "Disconnected"))
+                        extra = ": connecting to a PLAINTEXT broker listener?";
+
                 rd_kafka_broker_fail(rkb, LOG_ERR, RD_KAFKA_RESP_ERR__SSL,
                                      "SSL handshake failed: %s%s", errstr,
-                                     strstr(errstr, "unexpected message") ?
-                                     ": client authentication might be "
-                                     "required (see broker log)" : "");
+                                     extra);
                 return -1;
         }
 
@@ -643,28 +691,48 @@ static X509 *rd_kafka_ssl_X509_from_string (rd_kafka_t *rk, const char *str) {
 }
 
 
-#if _MSC_VER
+#ifdef _WIN32
 
 /**
- * @brief Attempt load CA certificates from the Windows Certificate Root store.
+ * @brief Attempt load CA certificates from a Windows Certificate store.
  */
-static int rd_kafka_ssl_win_load_root_certs (rd_kafka_t *rk, SSL_CTX *ctx) {
+static int rd_kafka_ssl_win_load_cert_store (rd_kafka_t *rk, SSL_CTX *ctx,
+                                             const char *store_name) {
         HCERTSTORE w_store;
         PCCERT_CONTEXT w_cctx = NULL;
         X509_STORE *store;
         int fail_cnt = 0, cnt = 0;
         char errstr[256];
+        wchar_t *wstore_name;
+        size_t wsize = 0;
+        errno_t werr;
+
+        /* Convert store_name to wide-char */
+        werr = mbstowcs_s(&wsize, NULL, 0, store_name, strlen(store_name));
+        if (werr || wsize < 2 || wsize > 1000) {
+                rd_kafka_log(rk, LOG_ERR, "CERTSTORE",
+                             "Invalid Windows certificate store name: %.*s%s",
+                             30, store_name,
+                             wsize < 2 ? " (empty)" : " (truncated)");
+                return -1;
+        }
+        wstore_name = rd_alloca(sizeof(*wstore_name) * wsize);
+        werr = mbstowcs_s(NULL, wstore_name, wsize, store_name,
+                          strlen(store_name));
+        rd_assert(!werr);
 
         w_store = CertOpenStore(CERT_STORE_PROV_SYSTEM,
                                 0,
                                 0,
-                                CERT_SYSTEM_STORE_CURRENT_USER,
-                                L"Root");
+                                CERT_SYSTEM_STORE_CURRENT_USER|
+                                CERT_STORE_READONLY_FLAG|
+                                CERT_STORE_OPEN_EXISTING_FLAG,
+                                wstore_name);
         if (!w_store) {
-                rd_kafka_dbg(rk, SECURITY, "CERTROOT",
+                rd_kafka_log(rk, LOG_ERR, "CERTSTORE",
                              "Failed to open Windows certificate "
-                             "Root store: %s: "
-                             "falling back to OpenSSL default CA paths",
+                             "%s store: %s",
+                             store_name,
                              rd_strerror_w32(GetLastError(), errstr,
                                              sizeof(errstr)));
                 return -1;
@@ -677,8 +745,6 @@ static int rd_kafka_ssl_win_load_root_certs (rd_kafka_t *rk, SSL_CTX *ctx) {
         while ((w_cctx = CertEnumCertificatesInStore(w_store, w_cctx))) {
                 X509 *x509;
 
-                cnt++;
-
                 /* Parse Windows cert: DER -> X.509 */
                 x509 = d2i_X509(NULL,
                                 (const unsigned char **)&w_cctx->pbCertEncoded,
@@ -690,7 +756,9 @@ static int rd_kafka_ssl_win_load_root_certs (rd_kafka_t *rk, SSL_CTX *ctx) {
 
                 /* Add cert to OpenSSL's trust store */
                 if (!X509_STORE_add_cert(store, x509))
-                        fail_cnt++;
+                    fail_cnt++;
+                else
+                    cnt++;
 
                 X509_free(x509);
         }
@@ -700,15 +768,175 @@ static int rd_kafka_ssl_win_load_root_certs (rd_kafka_t *rk, SSL_CTX *ctx) {
 
         CertCloseStore(w_store, 0);
 
-        rd_kafka_dbg(rk, SECURITY, "CERTROOT",
-                     "%d/%d certificate(s) successfully added from "
-                     "Windows Certificate Root store",
-                     cnt - fail_cnt, cnt);
+        rd_kafka_dbg(rk, SECURITY, "CERTSTORE",
+                     "%d certificate(s) successfully added from "
+                     "Windows Certificate %s store, %d failed",
+                     cnt, store_name, fail_cnt);
 
-        return cnt - fail_cnt == 0 ? -1 : 0;
+        if (cnt == 0 && fail_cnt > 0)
+            return -1;
+
+        return cnt;
+}
+
+/**
+ * @brief Load certs from the configured CSV list of Windows Cert stores.
+ *
+ * @returns the number of successfully loaded certificates, or -1 on error.
+ */
+static int rd_kafka_ssl_win_load_cert_stores (rd_kafka_t *rk,
+                                              SSL_CTX *ctx,
+                                              const char *store_names) {
+        char *s;
+        int cert_cnt = 0, fail_cnt = 0;
+
+        if (!store_names || !*store_names)
+                return 0;
+
+        rd_strdupa(&s, store_names);
+
+        /* Parse CSV list ("Root,CA, , ,Something") and load
+         * each store in order. */
+        while (*s) {
+                char *t;
+                const char *store_name;
+                int r;
+
+                while (isspace((int)*s) || *s == ',')
+                        s++;
+
+                if (!*s)
+                        break;
+
+                store_name = s;
+
+                t = strchr(s, (int)',');
+                if (t) {
+                        *t = '\0';
+                        s = t+1;
+                        for (; t >= store_name && isspace((int)*t) ; t--)
+                                     *t = '\0';
+                } else {
+                        s = "";
+                }
+
+                r = rd_kafka_ssl_win_load_cert_store(rk, ctx, store_name);
+                if (r != -1)
+                        cert_cnt += r;
+                else
+                        fail_cnt++;
+        }
+
+        if (cert_cnt == 0 && fail_cnt > 0)
+                return -1;
+
+        return cert_cnt;
 }
 #endif /* MSC_VER */
 
+
+
+/**
+ * @brief Probe for the system's CA certificate location and if found set it
+ *        on the \p CTX.
+ *
+ * @returns 0 if CA location was set, else -1.
+ */
+static int rd_kafka_ssl_probe_and_set_default_ca_location (rd_kafka_t *rk,
+                                                           SSL_CTX *ctx) {
+#if _WIN32
+        /* No standard location on Windows, CA certs are in the ROOT store. */
+        return -1;
+#else
+        /* The probe paths are based on:
+         * https://www.happyassassin.net/posts/2015/01/12/a-note-about-ssltls-trusted-certificate-stores-and-platforms/
+         * Golang's crypto probing paths:
+         *   https://golang.org/search?q=certFiles   and certDirectories
+         */
+        static const char *paths[] = {
+                "/etc/pki/tls/certs/ca-bundle.crt",
+                "/etc/ssl/certs/ca-bundle.crt",
+                "/etc/pki/tls/certs/ca-bundle.trust.crt",
+                "/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem",
+
+                "/etc/ssl/ca-bundle.pem",
+                "/etc/pki/tls/cacert.pem",
+                "/etc/ssl/cert.pem",
+                "/etc/ssl/cacert.pem",
+
+                "/etc/certs/ca-certificates.crt",
+                "/etc/ssl/certs/ca-certificates.crt",
+
+                "/etc/ssl/certs",
+
+                "/usr/local/etc/ssl/cert.pem",
+                "/usr/local/etc/ssl/cacert.pem",
+
+                "/usr/local/etc/ssl/certs/cert.pem",
+                "/usr/local/etc/ssl/certs/cacert.pem",
+
+                /* BSD */
+                "/usr/local/share/certs/ca-root-nss.crt",
+                "/etc/openssl/certs/ca-certificates.crt",
+#ifdef __APPLE__
+                "/private/etc/ssl/cert.pem",
+                "/private/etc/ssl/certs",
+                "/usr/local/etc/openssl@1.1/cert.pem",
+                "/usr/local/etc/openssl@1.0/cert.pem",
+                "/usr/local/etc/openssl/certs",
+                "/System/Library/OpenSSL",
+#endif
+#ifdef _AIX
+                "/var/ssl/certs/ca-bundle.crt",
+#endif
+                NULL,
+        };
+        const char *path = NULL;
+        int i;
+
+        for (i = 0 ; (path = paths[i]) ; i++) {
+                struct stat st;
+                rd_bool_t is_dir;
+                int r;
+
+                if (stat(path, &st) != 0)
+                        continue;
+
+                is_dir = S_ISDIR(st.st_mode);
+
+                if (is_dir && rd_kafka_dir_is_empty(path))
+                        continue;
+
+                rd_kafka_dbg(rk, SECURITY, "CACERTS",
+                             "Setting default CA certificate location "
+                             "to %s, override with ssl.ca.location", path);
+
+                r = SSL_CTX_load_verify_locations(ctx,
+                                                  is_dir ? NULL : path,
+                                                  is_dir ? path : NULL);
+                if (r != 1) {
+                        char errstr[512];
+                        /* Read error and clear the error stack */
+                        rd_kafka_ssl_error(rk, NULL, errstr, sizeof(errstr));
+                        rd_kafka_dbg(rk, SECURITY, "CACERTS",
+                                     "Failed to set default CA certificate "
+                                     "location to %s %s: %s: skipping",
+                                     is_dir ? "directory" : "file", path,
+                                     errstr);
+                        continue;
+                }
+
+                return 0;
+        }
+
+        rd_kafka_dbg(rk, SECURITY, "CACERTS",
+                     "Unable to find any standard CA certificate"
+                     "paths: is the ca-certificates package installed?");
+        return -1;
+#endif
+}
+
+
 /**
  * @brief Registers certificates, keys, etc, on the SSL_CTX
  *
@@ -733,7 +961,8 @@ static int rd_kafka_ssl_set_certs (rd_kafka_t *rk, SSL_CTX *ctx,
                 /* OpenSSL takes ownership of the store */
                 rk->rk_conf.ssl.ca->store = NULL;
 
-        } else if (rk->rk_conf.ssl.ca_location) {
+        } else if (rk->rk_conf.ssl.ca_location &&
+                   strcmp(rk->rk_conf.ssl.ca_location, "probe")) {
                 /* CA certificate location, either file or directory. */
                 int is_dir = rd_kafka_path_is_dir(rk->rk_conf.ssl.ca_location);
 
@@ -757,21 +986,59 @@ static int rd_kafka_ssl_set_certs (rd_kafka_t *rk, SSL_CTX *ctx,
                 }
 
         } else {
-#if _MSC_VER
+#ifdef _WIN32
                 /* Attempt to load CA root certificates from the
-                 * Windows crypto Root cert store. */
-                r = rd_kafka_ssl_win_load_root_certs(rk, ctx);
+                 * configured Windows certificate stores. */
+                r = rd_kafka_ssl_win_load_cert_stores(
+                        rk, ctx, rk->rk_conf.ssl.ca_cert_stores);
+                if (r == 0) {
+                        rd_kafka_log(rk, LOG_NOTICE, "CERTSTORE",
+                                "No CA certificates loaded from "
+                                "Windows certificate stores: "
+                                "falling back to default OpenSSL CA paths");
+                        r = -1;
+                } else if (r == -1)
+                        rd_kafka_log(rk, LOG_NOTICE, "CERTSTORE",
+                                "Failed to load CA certificates from "
+                                "Windows certificate stores: "
+                                "falling back to default OpenSSL CA paths");
 #else
                 r = -1;
 #endif
+
+                if ((rk->rk_conf.ssl.ca_location &&
+                     !strcmp(rk->rk_conf.ssl.ca_location, "probe"))
+#if WITH_STATIC_LIB_libcrypto
+                    || r == -1
+#endif
+                        ) {
+                        /* If OpenSSL was linked statically there is a risk
+                         * that the system installed CA certificate path
+                         * doesn't match the cert path of OpenSSL.
+                         * To circumvent this we check for the existence
+                         * of standard CA certificate paths and use the
+                         * first one that is found.
+                         * Ignore failures. */
+                        r = rd_kafka_ssl_probe_and_set_default_ca_location(
+                                rk, ctx);
+                }
+
                 if (r == -1) {
-                        /* Use default CA certificate paths: ignore failures */
+                        /* Use default CA certificate paths from linked OpenSSL:
+                         * ignore failures */
+
                         r = SSL_CTX_set_default_verify_paths(ctx);
-                        if (r != 1)
+                        if (r != 1) {
+                                char errstr2[512];
+                                /* Read error and clear the error stack. */
+                                rd_kafka_ssl_error(rk, NULL,
+                                                   errstr2, sizeof(errstr2));
                                 rd_kafka_dbg(
                                         rk, SECURITY, "SSL",
                                         "SSL_CTX_set_default_verify_paths() "
-                                        "failed: ignoring");
+                                        "failed: %s: ignoring", errstr2);
+                        }
+                        r = 0;
                 }
         }
 
@@ -1032,16 +1299,25 @@ void rd_kafka_ssl_ctx_term (rd_kafka_t *rk) {
 int rd_kafka_ssl_ctx_init (rd_kafka_t *rk, char *errstr, size_t errstr_size) {
         int r;
         SSL_CTX *ctx;
+        const char *linking =
+#if WITH_STATIC_LIB_libcrypto
+                "statically linked "
+#else
+                ""
+#endif
+                ;
 
 #if OPENSSL_VERSION_NUMBER >= 0x10100000
-        rd_kafka_dbg(rk, SECURITY, "OPENSSL", "Using OpenSSL version %s "
+        rd_kafka_dbg(rk, SECURITY, "OPENSSL", "Using %sOpenSSL version %s "
                      "(0x%lx, librdkafka built with 0x%lx)",
+                     linking,
                      OpenSSL_version(OPENSSL_VERSION),
                      OpenSSL_version_num(),
                      OPENSSL_VERSION_NUMBER);
 #else
-        rd_kafka_dbg(rk, SECURITY, "OPENSSL", "librdkafka built with OpenSSL "
-                     "version 0x%lx", OPENSSL_VERSION_NUMBER);
+        rd_kafka_dbg(rk, SECURITY, "OPENSSL",
+                     "librdkafka built with %sOpenSSL version 0x%lx",
+                     linking, OPENSSL_VERSION_NUMBER);
 #endif
 
         if (errstr_size > 0)
@@ -1146,7 +1422,7 @@ rd_kafka_transport_ssl_lock_cb (int mode, int i, const char *file, int line) {
 #endif
 
 static RD_UNUSED unsigned long rd_kafka_transport_ssl_threadid_cb (void) {
-#ifdef _MSC_VER
+#ifdef _WIN32
         /* Windows makes a distinction between thread handle
          * and thread id, which means we can't use the
          * thrd_current() API that returns the handle. */
