@@ -29,7 +29,7 @@
 #ifndef _RDKAFKA_TRANSPORT_H_
 #define _RDKAFKA_TRANSPORT_H_
 
-#ifndef _MSC_VER
+#ifndef _WIN32
 #include <poll.h>
 #endif
 
@@ -54,12 +54,19 @@ void rd_kafka_transport_request_sent (rd_kafka_broker_t *rkb,
 int rd_kafka_transport_framed_recv (rd_kafka_transport_t *rktrans,
                                     rd_kafka_buf_t **rkbufp,
                                     char *errstr, size_t errstr_size);
+
+rd_kafka_transport_t *rd_kafka_transport_new (rd_kafka_broker_t *rkb,
+                                              rd_socket_t s,
+                                              char *errstr,
+                                              size_t errstr_size);
 struct rd_kafka_broker_s;
 rd_kafka_transport_t *rd_kafka_transport_connect(struct rd_kafka_broker_s *rkb, const rd_sockaddr_inx_t *sinx,
                                                  char *errstr, size_t errstr_size);
 void rd_kafka_transport_connect_done (rd_kafka_transport_t *rktrans,
 				      char *errstr);
 
+void rd_kafka_transport_post_connect_setup (rd_kafka_transport_t *rktrans);
+
 void rd_kafka_transport_close(rd_kafka_transport_t *rktrans);
 void rd_kafka_transport_poll_set(rd_kafka_transport_t *rktrans, int event);
 void rd_kafka_transport_poll_clear(rd_kafka_transport_t *rktrans, int event);
