@@ -67,7 +67,7 @@
 #define le32toh letoh32
 #endif
 
-#if defined(__WIN32__) && defined(SG)
+#if !defined(__MINGW32__) && defined(__WIN32__) && defined(SG)
 struct iovec {
 	void *iov_base;	/* Pointer to data.  */
 	size_t iov_len;	/* Length of data.  */
