@@ -108,7 +108,7 @@ int cnd_timedwait_abs (cnd_t *cnd, mtx_t *mtx, const struct timespec *tspec) {
  * @name Read-write locks
  * @{
  */
-#ifndef _MSC_VER
+#ifndef _WIN32
 int rwlock_init (rwlock_t *rwl) {
         int r = pthread_rwlock_init(rwl, NULL);
         if (r) {
