@@ -44,12 +44,16 @@ static void do_test_reset (const char *topic, int partition,
 	rd_kafka_t *rk;
 	rd_kafka_topic_t *rkt;
 	int eofcnt = 0, msgcnt = 0, errcnt = 0;
+        rd_kafka_conf_t *conf;
 
 	TEST_SAY("Test auto.offset.reset=%s, "
 		 "expect %d msgs, %d EOFs, %d errors\n",
 		 reset, exp_msgcnt, exp_eofcnt, exp_errcnt);
 
-	rk = test_create_consumer(NULL, NULL, NULL, NULL);
+        test_conf_init(&conf, NULL, 60);
+        test_conf_set(conf, "enable.partition.eof", "true");
+
+	rk = test_create_consumer(NULL, NULL, conf, NULL);
 	rkt = test_create_topic_object(rk, topic, "auto.offset.reset", reset,
 				       NULL);
 
@@ -57,7 +61,7 @@ static void do_test_reset (const char *topic, int partition,
 	while (1) {
 		rd_kafka_message_t *rkm;
 
-		rkm = rd_kafka_consume(rkt, partition, 1000*10);
+		rkm = rd_kafka_consume(rkt, partition, tmout_multip(1000*10));
 		if (!rkm)
 			TEST_FAIL("%s: no message for 10s: "
 				  "%d/%d messages, %d/%d EOFs, %d/%d errors\n",
@@ -112,7 +116,7 @@ static void do_test_reset (const char *topic, int partition,
 int main_0034_offset_reset (int argc, char **argv) {
 	const char *topic = test_mk_topic_name(__FUNCTION__, 1);
 	const int partition = 0;
-	const int msgcnt = 100;
+        const int msgcnt = test_quick ? 20 : 100;
 
 	/* Produce messages */
 	test_produce_msgs_easy(topic, 0, partition, msgcnt);
