@@ -50,7 +50,7 @@ int main_0038_performance (int argc, char **argv) {
 	rd_kafka_t *rk;
 	rd_kafka_topic_t *rkt;
 	test_timing_t t_create, t_produce, t_consume;
-	int totsize = 1024*1024*128;
+	int totsize = 1024 * 1024 * (test_quick ? 8 : 128);
 	int msgcnt;
 
 	if (!strcmp(test_mode, "valgrind") || !strcmp(test_mode, "helgrind") ||
@@ -63,7 +63,7 @@ int main_0038_performance (int argc, char **argv) {
 		 msgcnt, (int)msgsize, topic, partition);
 	testid = test_id_generate();
 	test_conf_init(&conf, NULL, 120);
-	rd_kafka_conf_set_dr_cb(conf, test_dr_cb);
+	rd_kafka_conf_set_dr_msg_cb(conf, test_dr_msg_cb);
 	test_conf_set(conf, "queue.buffering.max.messages", "10000000");
         test_conf_set(conf, "linger.ms", "100");
 	rk = test_create_handle(RD_KAFKA_PRODUCER, conf);
