@@ -43,8 +43,8 @@ static size_t avail_broker_cnt;
 static void do_test_CreateTopics (const char *what,
                                   rd_kafka_t *rk, rd_kafka_queue_t *useq,
                                   int op_timeout, rd_bool_t validate_only) {
-        rd_kafka_queue_t *q = useq ? useq : rd_kafka_queue_new(rk);
-#define MY_NEW_TOPICS_CNT 6
+        rd_kafka_queue_t *q;
+#define MY_NEW_TOPICS_CNT 7
         char *topics[MY_NEW_TOPICS_CNT];
         rd_kafka_NewTopic_t *new_topics[MY_NEW_TOPICS_CNT];
         rd_kafka_AdminOptions_t *options = NULL;
@@ -69,25 +69,29 @@ static void do_test_CreateTopics (const char *what,
         int num_replicas = (int)avail_broker_cnt;
         int32_t *replicas;
 
+        SUB_TEST_QUICK("%s CreateTopics with %s, "
+                       "op_timeout %d, validate_only %d",
+                       rd_kafka_name(rk), what, op_timeout, validate_only);
+
+        q = useq ? useq : rd_kafka_queue_new(rk);
+
         /* Set up replicas */
         replicas = rd_alloca(sizeof(*replicas) * num_replicas);
         for (i = 0 ; i < num_replicas ; i++)
                 replicas[i] = avail_brokers[i];
 
-        TEST_SAY(_C_MAG "[ %s CreateTopics with %s, "
-                 "op_timeout %d, validate_only %d ]\n",
-                 rd_kafka_name(rk), what, op_timeout, validate_only);
-
         /**
          * Construct NewTopic array with different properties for
          * different partitions.
          */
         for (i = 0 ; i < MY_NEW_TOPICS_CNT ; i++) {
                 char *topic = rd_strdup(test_mk_topic_name(__FUNCTION__, 1));
-                int num_parts = i * 7 + 1;
+                int use_defaults = i == 6 &&
+                        test_broker_version >= TEST_BRKVER(2,4,0,0);
+                int num_parts = !use_defaults ? (i * 7 + 1) : -1;
                 int set_config = (i & 1);
                 int add_invalid_config = (i == 1);
-                int set_replicas = !(i % 3);
+                int set_replicas = !use_defaults && !(i % 3);
                 rd_kafka_resp_err_t this_exp_err = RD_KAFKA_RESP_ERR_NO_ERROR;
 
                 topics[i] = topic;
@@ -120,11 +124,12 @@ static void do_test_CreateTopics (const char *what,
                         this_exp_err = RD_KAFKA_RESP_ERR_INVALID_CONFIG;
                 }
 
-                TEST_SAY("Expected result for topic #%d: %s "
+                TEST_SAY("Expecting result for topic #%d: %s "
                          "(set_config=%d, add_invalid_config=%d, "
-                         "set_replicas=%d)\n",
+                         "set_replicas=%d, use_defaults=%d)\n",
                          i, rd_kafka_err2name(this_exp_err),
-                         set_config, add_invalid_config, set_replicas);
+                         set_config, add_invalid_config, set_replicas,
+                         use_defaults);
 
                 if (set_replicas) {
                         int32_t p;
@@ -274,7 +279,10 @@ static void do_test_CreateTopics (const char *what,
         if (!useq)
                 rd_kafka_queue_destroy(q);
 
+        TEST_LATER_CHECK();
 #undef MY_NEW_TOPICS_CNT
+
+        SUB_TEST_PASS();
 }
 
 
@@ -288,7 +296,7 @@ static void do_test_CreateTopics (const char *what,
 static void do_test_DeleteTopics (const char *what,
                                   rd_kafka_t *rk, rd_kafka_queue_t *useq,
                                   int op_timeout) {
-        rd_kafka_queue_t *q = useq ? useq : rd_kafka_queue_new(rk);
+        rd_kafka_queue_t *q;
         const int skip_topic_cnt = 2;
 #define MY_DEL_TOPICS_CNT 9
         char *topics[MY_DEL_TOPICS_CNT];
@@ -313,8 +321,10 @@ static void do_test_DeleteTopics (const char *what,
         size_t restopic_cnt;
         int metadata_tmout;
 
-        TEST_SAY(_C_MAG "[ %s DeleteTopics with %s, op_timeout %d ]\n",
-                 rd_kafka_name(rk), what, op_timeout);
+        SUB_TEST_QUICK("%s DeleteTopics with %s, op_timeout %d",
+                       rd_kafka_name(rk), what, op_timeout);
+
+        q = useq ? useq : rd_kafka_queue_new(rk);
 
         /**
          * Construct DeleteTopic array
@@ -463,7 +473,10 @@ static void do_test_DeleteTopics (const char *what,
         if (!useq)
                 rd_kafka_queue_destroy(q);
 
+        TEST_LATER_CHECK();
 #undef MY_DEL_TOPICS_CNT
+
+        SUB_TEST_PASS();
 }
 
 
@@ -476,7 +489,7 @@ static void do_test_DeleteTopics (const char *what,
 static void do_test_CreatePartitions (const char *what,
                                       rd_kafka_t *rk, rd_kafka_queue_t *useq,
                                       int op_timeout) {
-        rd_kafka_queue_t *q = useq ? useq : rd_kafka_queue_new(rk);
+        rd_kafka_queue_t *q;
 #define MY_CRP_TOPICS_CNT 9
         char *topics[MY_CRP_TOPICS_CNT];
         rd_kafka_NewTopic_t *new_topics[MY_CRP_TOPICS_CNT];
@@ -493,8 +506,10 @@ static void do_test_CreatePartitions (const char *what,
         int metadata_tmout;
         int num_replicas = (int)avail_broker_cnt;
 
-        TEST_SAY(_C_MAG "[ %s CreatePartitions with %s, op_timeout %d ]\n",
-                 rd_kafka_name(rk), what, op_timeout);
+        SUB_TEST_QUICK("%s CreatePartitions with %s, op_timeout %d",
+                       rd_kafka_name(rk), what, op_timeout);
+
+        q = useq ? useq : rd_kafka_queue_new(rk);
 
         /* Set up two expected partitions with different replication sets
          * so they can be matched by the metadata checker later.
@@ -659,7 +674,10 @@ static void do_test_CreatePartitions (const char *what,
         if (!useq)
                 rd_kafka_queue_destroy(q);
 
+        TEST_LATER_CHECK();
 #undef MY_CRP_TOPICS_CNT
+
+        SUB_TEST_PASS();
 }
 
 
@@ -727,6 +745,8 @@ static void do_test_AlterConfigs (rd_kafka_t *rk, rd_kafka_queue_t *rkqu) {
         int i;
         int fails = 0;
 
+        SUB_TEST_QUICK();
+
         /*
          * Only create one topic, the others will be non-existent.
          */
@@ -735,6 +755,8 @@ static void do_test_AlterConfigs (rd_kafka_t *rk, rd_kafka_queue_t *rkqu) {
 
         test_CreateTopics_simple(rk, NULL, topics, 1, 1, NULL);
 
+        test_wait_topic_exists(rk, topics[0], 10000);
+
         /*
          * ConfigResource #0: valid topic config
          */
@@ -788,7 +810,10 @@ static void do_test_AlterConfigs (rd_kafka_t *rk, rd_kafka_queue_t *rkqu) {
                                                  "12345");
         TEST_ASSERT(!err, "%s", rd_kafka_err2str(err));
 
-        exp_err[ci] = RD_KAFKA_RESP_ERR_UNKNOWN;
+        if (test_broker_version >= TEST_BRKVER(2, 7, 0, 0))
+                exp_err[ci] = RD_KAFKA_RESP_ERR_UNKNOWN_TOPIC_OR_PART;
+        else
+                exp_err[ci] = RD_KAFKA_RESP_ERR_UNKNOWN;
         ci++;
 
 
@@ -893,7 +918,10 @@ static void do_test_AlterConfigs (rd_kafka_t *rk, rd_kafka_queue_t *rkqu) {
 
         rd_kafka_ConfigResource_destroy_array(configs, ci);
 
+        TEST_LATER_CHECK();
 #undef MY_CONFRES_CNT
+
+        SUB_TEST_PASS();
 }
 
 
@@ -919,6 +947,8 @@ static void do_test_DescribeConfigs (rd_kafka_t *rk, rd_kafka_queue_t *rkqu) {
         int fails = 0;
         int max_retry_describe = 3;
 
+        SUB_TEST_QUICK();
+
         /*
          * Only create one topic, the others will be non-existent.
          */
@@ -1083,7 +1113,10 @@ static void do_test_DescribeConfigs (rd_kafka_t *rk, rd_kafka_queue_t *rkqu) {
 
         rd_kafka_ConfigResource_destroy_array(configs, ci);
 
+        TEST_LATER_CHECK();
 #undef MY_CONFRES_CNT
+
+        SUB_TEST_PASS();
 }
 
 
@@ -1099,14 +1132,14 @@ static void do_test_unclean_destroy (rd_kafka_type_t cltype, int with_mainq) {
         rd_kafka_NewTopic_t *topic;
         test_timing_t t_destroy;
 
+        SUB_TEST_QUICK("Test unclean destroy using %s",
+                       with_mainq ? "mainq" : "tempq");
+
         test_conf_init(&conf, NULL, 0);
 
         rk = rd_kafka_new(cltype, conf, errstr, sizeof(errstr));
         TEST_ASSERT(rk, "kafka_new(%d): %s", cltype, errstr);
 
-        TEST_SAY(_C_MAG "[ Test unclean destroy for %s using %s]\n", rd_kafka_name(rk),
-                 with_mainq ? "mainq" : "tempq");
-
         if (with_mainq)
                 q = rd_kafka_queue_get_main(rk);
         else
@@ -1126,12 +1159,735 @@ static void do_test_unclean_destroy (rd_kafka_type_t cltype, int with_mainq) {
         rd_kafka_destroy(rk);
         TIMING_STOP(&t_destroy);
 
+        SUB_TEST_PASS();
+
         /* Restore timeout */
-        test_timeout_set(60);;
+        test_timeout_set(60);
 }
 
 
 
+
+/**
+  * @brief Test deletion of records
+  *
+  *
+  */
+static void do_test_DeleteRecords (const char *what,
+                                   rd_kafka_t *rk, rd_kafka_queue_t *useq,
+                                   int op_timeout) {
+        rd_kafka_queue_t *q;
+        rd_kafka_AdminOptions_t *options = NULL;
+        rd_kafka_topic_partition_list_t *offsets = NULL;
+        rd_kafka_event_t *rkev = NULL;
+        rd_kafka_resp_err_t err;
+        char errstr[512];
+        const char *errstr2;
+#define MY_DEL_RECORDS_CNT 3
+        rd_kafka_topic_partition_list_t *results = NULL;
+        int i;
+        const int partitions_cnt = 3;
+        const int msgs_cnt = 100;
+        char *topics[MY_DEL_RECORDS_CNT];
+        rd_kafka_metadata_topic_t exp_mdtopics[MY_DEL_RECORDS_CNT] = {{0}};
+        int exp_mdtopic_cnt = 0;
+        test_timing_t timing;
+        rd_kafka_resp_err_t exp_err = RD_KAFKA_RESP_ERR_NO_ERROR;
+        rd_kafka_DeleteRecords_t *del_records;
+        const rd_kafka_DeleteRecords_result_t *res;
+
+        SUB_TEST_QUICK("%s DeleteRecords with %s, op_timeout %d",
+                       rd_kafka_name(rk), what, op_timeout);
+
+        q = useq ? useq : rd_kafka_queue_new(rk);
+
+        if (op_timeout != -1) {
+                options = rd_kafka_AdminOptions_new(
+                        rk, RD_KAFKA_ADMIN_OP_ANY);
+
+                err = rd_kafka_AdminOptions_set_operation_timeout(
+                        options, op_timeout, errstr, sizeof(errstr));
+                TEST_ASSERT(!err, "%s", rd_kafka_err2str(err));
+        }
+
+
+        for (i = 0 ; i < MY_DEL_RECORDS_CNT ; i++) {
+                char pfx[32];
+                char *topic;
+
+                rd_snprintf(pfx, sizeof(pfx), "DeleteRecords-topic%d", i);
+                topic = rd_strdup(test_mk_topic_name(pfx, 1));
+
+                topics[i] = topic;
+                exp_mdtopics[exp_mdtopic_cnt++].topic = topic;
+        }
+
+        /* Create the topics first. */
+        test_CreateTopics_simple(rk, NULL, topics,
+                                 MY_DEL_RECORDS_CNT,
+                                 partitions_cnt /*num_partitions*/,
+                                 NULL);
+
+        /* Verify that topics are reported by metadata */
+        test_wait_metadata_update(rk,
+                                  exp_mdtopics, exp_mdtopic_cnt,
+                                  NULL, 0,
+                                  15*1000);
+
+        /* Produce 100 msgs / partition */
+        for (i = 0 ; i < MY_DEL_RECORDS_CNT; i++ ) {
+                int32_t partition;
+                for (partition = 0 ; partition < partitions_cnt; partition++ ) {
+                        test_produce_msgs_easy(topics[i], 0, partition,
+                                               msgs_cnt);
+                }
+        }
+
+        offsets = rd_kafka_topic_partition_list_new(10);
+
+        /* Wipe all data from topic 0 */
+        for (i = 0 ; i < partitions_cnt; i++)
+                rd_kafka_topic_partition_list_add(offsets, topics[0], i)->
+                        offset = RD_KAFKA_OFFSET_END;
+
+        /* Wipe all data from partition 0 in topic 1 */
+        rd_kafka_topic_partition_list_add(offsets, topics[1], 0)->
+                offset = RD_KAFKA_OFFSET_END;
+
+        /* Wipe some data from partition 2 in topic 1 */
+        rd_kafka_topic_partition_list_add(offsets, topics[1], 2)->
+                offset = msgs_cnt / 2;
+
+        /* Not changing the offset (out of range) for topic 2 partition 0 */
+        rd_kafka_topic_partition_list_add(offsets, topics[2], 0);
+
+        /* Offset out of range for topic 2 partition 1 */
+        rd_kafka_topic_partition_list_add(offsets, topics[2], 1)->
+                offset = msgs_cnt + 1;
+
+        del_records = rd_kafka_DeleteRecords_new(offsets);
+
+        TIMING_START(&timing, "DeleteRecords");
+        TEST_SAY("Call DeleteRecords\n");
+        rd_kafka_DeleteRecords(rk, &del_records, 1, options, q);
+        TIMING_ASSERT_LATER(&timing, 0, 50);
+
+        rd_kafka_DeleteRecords_destroy(del_records);
+
+        TIMING_START(&timing, "DeleteRecords.queue_poll");
+
+        /* Poll result queue for DeleteRecords result.
+         * Print but otherwise ignore other event types
+         * (typically generic Error events). */
+        while (1) {
+                rkev = rd_kafka_queue_poll(q, tmout_multip(20*1000));
+                TEST_SAY("DeleteRecords: got %s in %.3fms\n",
+                         rd_kafka_event_name(rkev),
+                         TIMING_DURATION(&timing) / 1000.0f);
+                if (rkev == NULL)
+                        continue;
+                if (rd_kafka_event_error(rkev))
+                        TEST_SAY("%s: %s\n",
+                                 rd_kafka_event_name(rkev),
+                                 rd_kafka_event_error_string(rkev));
+
+                if (rd_kafka_event_type(rkev) ==
+                    RD_KAFKA_EVENT_DELETERECORDS_RESULT) {
+                        break;
+                }
+
+                rd_kafka_event_destroy(rkev);
+        }
+        /* Convert event to proper result */
+        res = rd_kafka_event_DeleteRecords_result(rkev);
+        TEST_ASSERT(res, "expected DeleteRecords_result, not %s",
+                    rd_kafka_event_name(rkev));
+
+        /* Expecting error */
+        err = rd_kafka_event_error(rkev);
+        errstr2 = rd_kafka_event_error_string(rkev);
+        TEST_ASSERT(err == exp_err,
+                    "expected DeleteRecords to return %s, not %s (%s)",
+                    rd_kafka_err2str(exp_err),
+                    rd_kafka_err2str(err),
+                    err ? errstr2 : "n/a");
+
+        TEST_SAY("DeleteRecords: returned %s (%s)\n",
+                 rd_kafka_err2str(err), err ? errstr2 : "n/a");
+
+        results = rd_kafka_topic_partition_list_copy(
+                rd_kafka_DeleteRecords_result_offsets(res));
+
+        /* Sort both input and output list */
+        rd_kafka_topic_partition_list_sort(offsets, NULL, NULL);
+        rd_kafka_topic_partition_list_sort(results, NULL, NULL);
+
+        TEST_SAY("Input partitions:\n");
+        test_print_partition_list(offsets);
+        TEST_SAY("Result partitions:\n");
+        test_print_partition_list(results);
+
+        TEST_ASSERT(offsets->cnt == results->cnt,
+                    "expected DeleteRecords_result_offsets to return %d items, "
+                    "not %d",
+                    offsets->cnt,
+                    results->cnt);
+
+        for (i = 0 ; i < results->cnt ; i++) {
+                const rd_kafka_topic_partition_t *input =&offsets->elems[i];
+                const rd_kafka_topic_partition_t *output = &results->elems[i];
+                int64_t expected_offset = input->offset;
+                rd_kafka_resp_err_t expected_err = 0;
+
+                if (expected_offset == RD_KAFKA_OFFSET_END)
+                        expected_offset = msgs_cnt;
+
+                /* Expect Offset out of range error */
+                if (input->offset < RD_KAFKA_OFFSET_END ||
+                    input->offset > msgs_cnt)
+                        expected_err = 1;
+
+                TEST_SAY("DeleteRecords Returned %s for %s [%"PRId32"] "
+                         "low-watermark = %d\n",
+                         rd_kafka_err2name(output->err),
+                         output->topic,
+                         output->partition,
+                         (int)output->offset);
+
+                if (strcmp(output->topic, input->topic))
+                        TEST_FAIL_LATER("Result order mismatch at #%d: "
+                                        "expected topic %s, got %s",
+                                        i,
+                                        input->topic,
+                                        output->topic);
+
+                if (output->partition != input->partition)
+                        TEST_FAIL_LATER("Result order mismatch at #%d: "
+                                        "expected partition %d, got %d",
+                                        i,
+                                        input->partition,
+                                        output->partition);
+
+                if (output->err != expected_err)
+                        TEST_FAIL_LATER("%s [%"PRId32"]: "
+                                        "expected error code %d (%s), "
+                                        "got %d (%s)",
+                                        output->topic,
+                                        output->partition,
+                                        expected_err,
+                                        rd_kafka_err2str(expected_err),
+                                        output->err,
+                                        rd_kafka_err2str(output->err));
+
+                if (output->err == 0 && output->offset != expected_offset)
+                        TEST_FAIL_LATER("%s [%"PRId32"]: "
+                                        "expected offset %"PRId64", "
+                                        "got %"PRId64,
+                                        output->topic,
+                                        output->partition,
+                                        expected_offset,
+                                        output->offset);
+        }
+
+        /* Check watermarks for partitions */
+        for (i = 0 ; i < MY_DEL_RECORDS_CNT; i++ ) {
+                int32_t partition;
+                for (partition = 0 ; partition < partitions_cnt; partition++ ) {
+                        const rd_kafka_topic_partition_t *del =
+                                rd_kafka_topic_partition_list_find(
+                                        results, topics[i], partition);
+                        int64_t expected_low = 0;
+                        int64_t expected_high = msgs_cnt;
+                        int64_t low, high;
+
+                        if (del && del->err == 0) {
+                                expected_low = del->offset;
+                        }
+
+                        err = rd_kafka_query_watermark_offsets(
+                                rk, topics[i], partition,
+                                &low, &high, tmout_multip(10000));
+                        if (err)
+                                TEST_FAIL("query_watermark_offsets failed: "
+                                          "%s\n",
+                                          rd_kafka_err2str(err));
+
+                        if (low != expected_low)
+                                TEST_FAIL_LATER("For %s [%"PRId32"] expected "
+                                                "a low watermark of %"PRId64
+                                                ", got %"PRId64,
+                                                topics[i],
+                                                partition,
+                                                expected_low,
+                                                low);
+
+                        if (high != expected_high)
+                                TEST_FAIL_LATER("For %s [%"PRId32"] expected "
+                                                "a high watermark of %"PRId64
+                                                ", got %"PRId64,
+                                                topics[i],
+                                                partition,
+                                                expected_high,
+                                                high);
+                }
+        }
+
+        rd_kafka_event_destroy(rkev);
+
+        for (i = 0 ; i < MY_DEL_RECORDS_CNT ; i++)
+                rd_free(topics[i]);
+
+        if (results)
+                rd_kafka_topic_partition_list_destroy(results);
+
+        if (offsets)
+                rd_kafka_topic_partition_list_destroy(offsets);
+
+        if (options)
+                rd_kafka_AdminOptions_destroy(options);
+
+        if (!useq)
+                rd_kafka_queue_destroy(q);
+
+        TEST_LATER_CHECK();
+#undef MY_DEL_RECORDS_CNT
+
+        SUB_TEST_PASS();
+}
+
+/**
+  * @brief Test deletion of groups
+  *
+  *
+  */
+
+typedef struct expected_group_result {
+        char *group;
+        rd_kafka_resp_err_t err;
+} expected_group_result_t;
+
+static void do_test_DeleteGroups (const char *what,
+                                  rd_kafka_t *rk, rd_kafka_queue_t *useq,
+                                  int op_timeout) {
+        rd_kafka_queue_t *q;
+        rd_kafka_AdminOptions_t *options = NULL;
+        rd_kafka_event_t *rkev = NULL;
+        rd_kafka_resp_err_t err;
+        char errstr[512];
+        const char *errstr2;
+#define MY_DEL_GROUPS_CNT 4
+        int known_groups = MY_DEL_GROUPS_CNT - 1;
+        int i;
+        const int partitions_cnt = 1;
+        const int msgs_cnt = 100;
+        char *topic;
+        rd_kafka_metadata_topic_t exp_mdtopic = {0};
+        int64_t testid = test_id_generate();
+        test_timing_t timing;
+        rd_kafka_resp_err_t exp_err = RD_KAFKA_RESP_ERR_NO_ERROR;
+        const rd_kafka_group_result_t **results = NULL;
+        expected_group_result_t expected[MY_DEL_GROUPS_CNT] = {{0}};
+        rd_kafka_DeleteGroup_t *del_groups[MY_DEL_GROUPS_CNT];
+        const rd_kafka_DeleteGroups_result_t *res;
+
+        SUB_TEST_QUICK("%s DeleteGroups with %s, op_timeout %d",
+                       rd_kafka_name(rk), what, op_timeout);
+
+        q = useq ? useq : rd_kafka_queue_new(rk);
+
+        if (op_timeout != -1) {
+                options = rd_kafka_AdminOptions_new(
+                        rk, RD_KAFKA_ADMIN_OP_ANY);
+
+                err = rd_kafka_AdminOptions_set_operation_timeout(
+                        options, op_timeout, errstr, sizeof(errstr));
+                TEST_ASSERT(!err, "%s", rd_kafka_err2str(err));
+        }
+
+
+        topic = rd_strdup(test_mk_topic_name(__FUNCTION__, 1));
+        exp_mdtopic.topic = topic;
+
+        /* Create the topics first. */
+        test_CreateTopics_simple(rk, NULL, &topic, 1,
+                                 partitions_cnt,
+                                 NULL);
+
+        /* Verify that topics are reported by metadata */
+        test_wait_metadata_update(rk,
+                                  &exp_mdtopic, 1,
+                                  NULL, 0,
+                                  15*1000);
+
+        /* Produce 100 msgs */
+        test_produce_msgs_easy(topic, testid, 0, msgs_cnt);
+
+        for (i = 0; i < MY_DEL_GROUPS_CNT; i++) {
+                char *group = rd_strdup(test_mk_topic_name(__FUNCTION__, 1));
+                if (i < known_groups) {
+                        test_consume_msgs_easy(group, topic, testid, -1, msgs_cnt, NULL);
+                        expected[i].group = group;
+                        expected[i].err = RD_KAFKA_RESP_ERR_NO_ERROR;
+                } else {
+                        expected[i].group = group;
+                        expected[i].err = RD_KAFKA_RESP_ERR_GROUP_ID_NOT_FOUND;
+                }
+                del_groups[i] = rd_kafka_DeleteGroup_new(group);
+        }
+
+        TIMING_START(&timing, "DeleteGroups");
+        TEST_SAY("Call DeleteGroups\n");
+        rd_kafka_DeleteGroups(rk, del_groups, MY_DEL_GROUPS_CNT, options, q);
+        TIMING_ASSERT_LATER(&timing, 0, 50);
+
+        TIMING_START(&timing, "DeleteGroups.queue_poll");
+
+        /* Poll result queue for DeleteGroups result.
+         * Print but otherwise ignore other event types
+         * (typically generic Error events). */
+        while(1) {
+                rkev = rd_kafka_queue_poll(q, tmout_multip(20*1000));
+                TEST_SAY("DeleteGroups: got %s in %.3fms\n",
+                         rd_kafka_event_name(rkev),
+                         TIMING_DURATION(&timing) / 1000.0f);
+                if (rkev == NULL)
+                        continue;
+                if (rd_kafka_event_error(rkev))
+                        TEST_SAY("%s: %s\n",
+                                 rd_kafka_event_name(rkev),
+                                 rd_kafka_event_error_string(rkev));
+
+                if (rd_kafka_event_type(rkev) ==
+                    RD_KAFKA_EVENT_DELETEGROUPS_RESULT) {
+                        break;
+                }
+
+                rd_kafka_event_destroy(rkev);
+        }
+        /* Convert event to proper result */
+        res = rd_kafka_event_DeleteGroups_result(rkev);
+        TEST_ASSERT(res, "expected DeleteGroups_result, not %s",
+                    rd_kafka_event_name(rkev));
+
+        /* Expecting error */
+        err = rd_kafka_event_error(rkev);
+        errstr2 = rd_kafka_event_error_string(rkev);
+        TEST_ASSERT(err == exp_err,
+                    "expected DeleteGroups to return %s, not %s (%s)",
+                    rd_kafka_err2str(exp_err),
+                    rd_kafka_err2str(err),
+                    err ? errstr2 : "n/a");
+
+        TEST_SAY("DeleteGroups: returned %s (%s)\n",
+                 rd_kafka_err2str(err), err ? errstr2 : "n/a");
+
+        size_t cnt = 0;
+        results = rd_kafka_DeleteGroups_result_groups(res, &cnt);
+
+        TEST_ASSERT(MY_DEL_GROUPS_CNT == cnt,
+                    "expected DeleteGroups_result_groups to return %d items, not %"PRIusz,
+                    MY_DEL_GROUPS_CNT,
+                    cnt);
+
+        for (i = 0 ; i < MY_DEL_GROUPS_CNT ; i++) {
+                const expected_group_result_t *exp = &expected[i];
+                rd_kafka_resp_err_t exp_err = exp->err;
+                const rd_kafka_group_result_t *act = results[i];
+                rd_kafka_resp_err_t act_err = rd_kafka_error_code(rd_kafka_group_result_error(act));
+                TEST_ASSERT(strcmp(exp->group, rd_kafka_group_result_name(act)) == 0,
+                            "Result order mismatch at #%d: expected group name to be %s, not %s",
+                            i, exp->group, rd_kafka_group_result_name(act));
+                TEST_ASSERT(exp_err == act_err,
+                            "expected err=%d for group %s, not %d (%s)",
+                            exp_err,
+                            exp->group,
+                            act_err,
+                            rd_kafka_err2str(act_err));
+        }
+
+        rd_kafka_event_destroy(rkev);
+
+        for (i = 0 ; i < MY_DEL_GROUPS_CNT ; i++) {
+                rd_kafka_DeleteGroup_destroy(del_groups[i]);
+                rd_free(expected[i].group);
+        }
+
+        rd_free(topic);
+
+        if (options)
+                rd_kafka_AdminOptions_destroy(options);
+
+        if (!useq)
+                rd_kafka_queue_destroy(q);
+
+        TEST_LATER_CHECK();
+#undef MY_DEL_GROUPS_CNT
+
+        SUB_TEST_PASS();
+}
+
+
+/**
+  * @brief Test deletion of committed offsets.
+  *
+  *
+  */
+static void do_test_DeleteConsumerGroupOffsets (const char *what,
+                                                rd_kafka_t *rk,
+                                                rd_kafka_queue_t *useq,
+                                                int op_timeout,
+                                                rd_bool_t sub_consumer) {
+        rd_kafka_queue_t *q;
+        rd_kafka_AdminOptions_t *options = NULL;
+        rd_kafka_topic_partition_list_t *orig_offsets, *offsets,
+                *to_delete, *committed, *deleted, *subscription = NULL;
+        rd_kafka_event_t *rkev = NULL;
+        rd_kafka_resp_err_t err;
+        char errstr[512];
+        const char *errstr2;
+#define MY_TOPIC_CNT 3
+        int i;
+        const int partitions_cnt = 3;
+        char *topics[MY_TOPIC_CNT];
+        rd_kafka_metadata_topic_t exp_mdtopics[MY_TOPIC_CNT] = {{0}};
+        int exp_mdtopic_cnt = 0;
+        test_timing_t timing;
+        rd_kafka_resp_err_t exp_err = RD_KAFKA_RESP_ERR_NO_ERROR;
+        rd_kafka_DeleteConsumerGroupOffsets_t *cgoffsets;
+        const rd_kafka_DeleteConsumerGroupOffsets_result_t *res;
+        const rd_kafka_group_result_t **gres;
+        size_t gres_cnt;
+        rd_kafka_t *consumer;
+        char *groupid;
+
+        SUB_TEST_QUICK("%s DeleteConsumerGroupOffsets with %s, op_timeout %d%s",
+                       rd_kafka_name(rk), what, op_timeout,
+                       sub_consumer ? ", with subscribing consumer" : "");
+
+        if (sub_consumer)
+                exp_err = RD_KAFKA_RESP_ERR_GROUP_SUBSCRIBED_TO_TOPIC;
+
+        q = useq ? useq : rd_kafka_queue_new(rk);
+
+        if (op_timeout != -1) {
+                options = rd_kafka_AdminOptions_new(
+                        rk, RD_KAFKA_ADMIN_OP_ANY);
+
+                err = rd_kafka_AdminOptions_set_operation_timeout(
+                        options, op_timeout, errstr, sizeof(errstr));
+                TEST_ASSERT(!err, "%s", rd_kafka_err2str(err));
+        }
+
+
+        subscription = rd_kafka_topic_partition_list_new(MY_TOPIC_CNT);
+
+        for (i = 0 ; i < MY_TOPIC_CNT ; i++) {
+                char pfx[64];
+                char *topic;
+
+                rd_snprintf(pfx, sizeof(pfx),
+                            "DCGO-topic%d", i);
+                topic = rd_strdup(test_mk_topic_name(pfx, 1));
+
+                topics[i] = topic;
+                exp_mdtopics[exp_mdtopic_cnt++].topic = topic;
+
+                rd_kafka_topic_partition_list_add(subscription, topic,
+                                                  RD_KAFKA_PARTITION_UA);
+        }
+
+        groupid = topics[0];
+
+        /* Create the topics first. */
+        test_CreateTopics_simple(rk, NULL, topics, MY_TOPIC_CNT,
+                                 partitions_cnt, NULL);
+
+        /* Verify that topics are reported by metadata */
+        test_wait_metadata_update(rk,
+                                  exp_mdtopics, exp_mdtopic_cnt,
+                                  NULL, 0,
+                                  15*1000);
+
+        consumer = test_create_consumer(groupid, NULL, NULL, NULL);
+
+        if (sub_consumer) {
+                TEST_CALL_ERR__(rd_kafka_subscribe(consumer, subscription));
+                test_consumer_wait_assignment(consumer);
+        }
+
+        /* Commit some offsets */
+        orig_offsets = rd_kafka_topic_partition_list_new(MY_TOPIC_CNT * 2);
+        for (i = 0 ; i < MY_TOPIC_CNT * 2 ; i++)
+                rd_kafka_topic_partition_list_add(
+                        orig_offsets, topics[i/2],
+                        i % MY_TOPIC_CNT)->offset = (i+1)*10;
+
+        TEST_CALL_ERR__(rd_kafka_commit(consumer, orig_offsets, 0/*sync*/));
+
+        /* Verify committed offsets match */
+        committed = rd_kafka_topic_partition_list_copy(orig_offsets);
+        TEST_CALL_ERR__(rd_kafka_committed(consumer, committed,
+                                           tmout_multip(5*1000)));
+
+        if (test_partition_list_cmp(committed, orig_offsets)) {
+                TEST_SAY("commit() list:\n");
+                test_print_partition_list(orig_offsets);
+                TEST_SAY("committed() list:\n");
+                test_print_partition_list(committed);
+                TEST_FAIL("committed offsets don't match");
+        }
+
+        rd_kafka_topic_partition_list_destroy(committed);
+
+        /* Now delete second half of the commits */
+        offsets = rd_kafka_topic_partition_list_new(orig_offsets->cnt / 2);
+        to_delete = rd_kafka_topic_partition_list_new(orig_offsets->cnt / 2);
+        for (i = 0 ; i < orig_offsets->cnt ; i++) {
+                if (i < orig_offsets->cnt / 2)
+                        rd_kafka_topic_partition_list_add(
+                                offsets,
+                                orig_offsets->elems[i].topic,
+                                orig_offsets->elems[i].partition);
+                else {
+                        rd_kafka_topic_partition_list_add(
+                                to_delete,
+                                orig_offsets->elems[i].topic,
+                                orig_offsets->elems[i].partition);
+                        rd_kafka_topic_partition_list_add(
+                                offsets,
+                                orig_offsets->elems[i].topic,
+                                orig_offsets->elems[i].partition)->offset =
+                                RD_KAFKA_OFFSET_INVALID;
+                }
+
+        }
+
+        cgoffsets = rd_kafka_DeleteConsumerGroupOffsets_new(groupid, to_delete);
+
+        TIMING_START(&timing, "DeleteConsumerGroupOffsets");
+        TEST_SAY("Call DeleteConsumerGroupOffsets\n");
+        rd_kafka_DeleteConsumerGroupOffsets(rk, &cgoffsets, 1, options, q);
+        TIMING_ASSERT_LATER(&timing, 0, 50);
+
+        rd_kafka_DeleteConsumerGroupOffsets_destroy(cgoffsets);
+
+        TIMING_START(&timing, "DeleteConsumerGroupOffsets.queue_poll");
+        /* Poll result queue for DeleteConsumerGroupOffsets result.
+         * Print but otherwise ignore other event types
+         * (typically generic Error events). */
+        while (1) {
+                rkev = rd_kafka_queue_poll(q, tmout_multip(10*1000));
+                TEST_SAY("DeleteConsumerGroupOffsets: got %s in %.3fms\n",
+                         rd_kafka_event_name(rkev),
+                         TIMING_DURATION(&timing) / 1000.0f);
+                if (rkev == NULL)
+                        continue;
+                if (rd_kafka_event_error(rkev))
+                        TEST_SAY("%s: %s\n",
+                                 rd_kafka_event_name(rkev),
+                                 rd_kafka_event_error_string(rkev));
+
+                if (rd_kafka_event_type(rkev) ==
+                    RD_KAFKA_EVENT_DELETECONSUMERGROUPOFFSETS_RESULT)
+                        break;
+
+                rd_kafka_event_destroy(rkev);
+        }
+
+        /* Convert event to proper result */
+        res = rd_kafka_event_DeleteConsumerGroupOffsets_result(rkev);
+        TEST_ASSERT(res, "expected DeleteConsumerGroupOffsets_result, not %s",
+                    rd_kafka_event_name(rkev));
+
+        /* Expecting error */
+        err = rd_kafka_event_error(rkev);
+        errstr2 = rd_kafka_event_error_string(rkev);
+        TEST_ASSERT(!err,
+                    "expected DeleteConsumerGroupOffsets to succeed, "
+                    "got %s (%s)",
+                    rd_kafka_err2name(err),
+                    err ? errstr2 : "n/a");
+
+        TEST_SAY("DeleteConsumerGroupOffsets: returned %s (%s)\n",
+                 rd_kafka_err2str(err), err ? errstr2 : "n/a");
+
+        gres = rd_kafka_DeleteConsumerGroupOffsets_result_groups(res,
+                                                                 &gres_cnt);
+        TEST_ASSERT(gres && gres_cnt == 1,
+                    "expected gres_cnt == 1, not %"PRIusz, gres_cnt);
+
+        deleted = rd_kafka_topic_partition_list_copy(
+                rd_kafka_group_result_partitions(gres[0]));
+
+        if (test_partition_list_cmp(deleted, to_delete)) {
+                TEST_SAY("Result list:\n");
+                test_print_partition_list(deleted);
+                TEST_SAY("Partitions passed to DeleteConsumerGroupOffsets:\n");
+                test_print_partition_list(to_delete);
+                TEST_FAIL("deleted/requested offsets don't match");
+        }
+
+        /* Verify expected errors */
+        for (i = 0 ; i < deleted->cnt ; i++) {
+                TEST_ASSERT_LATER(deleted->elems[i].err == exp_err,
+                                  "Result %s [%"PRId32"] has error %s, "
+                                  "expected %s",
+                                  deleted->elems[i].topic,
+                                  deleted->elems[i].partition,
+                                  rd_kafka_err2name(deleted->elems[i].err),
+                                  rd_kafka_err2name(exp_err));
+        }
+
+        TEST_LATER_CHECK();
+
+        rd_kafka_topic_partition_list_destroy(deleted);
+        rd_kafka_topic_partition_list_destroy(to_delete);
+
+        rd_kafka_event_destroy(rkev);
+
+
+        /* Verify committed offsets match */
+        committed = rd_kafka_topic_partition_list_copy(orig_offsets);
+        TEST_CALL_ERR__(rd_kafka_committed(consumer, committed,
+                                           tmout_multip(5*1000)));
+
+        TEST_SAY("Original committed offsets:\n");
+        test_print_partition_list(orig_offsets);
+
+        TEST_SAY("Committed offsets after delete:\n");
+        test_print_partition_list(committed);
+
+        if (test_partition_list_cmp(committed, offsets)) {
+                TEST_SAY("expected list:\n");
+                test_print_partition_list(offsets);
+                TEST_SAY("committed() list:\n");
+                test_print_partition_list(committed);
+                TEST_FAIL("committed offsets don't match");
+        }
+
+        rd_kafka_topic_partition_list_destroy(committed);
+        rd_kafka_topic_partition_list_destroy(offsets);
+        rd_kafka_topic_partition_list_destroy(orig_offsets);
+        rd_kafka_topic_partition_list_destroy(subscription);
+
+        for (i = 0 ; i < MY_TOPIC_CNT ; i++)
+                rd_free(topics[i]);
+
+        rd_kafka_destroy(consumer);
+
+        if (options)
+                rd_kafka_AdminOptions_destroy(options);
+
+        if (!useq)
+                rd_kafka_queue_destroy(q);
+
+        TEST_LATER_CHECK();
+#undef MY_DEL_RECORDS_CNT
+
+        SUB_TEST_PASS();
+}
+
+
 static void do_test_apis (rd_kafka_type_t cltype) {
         rd_kafka_t *rk;
         rd_kafka_conf_t *conf;
@@ -1147,7 +1903,7 @@ static void do_test_apis (rd_kafka_type_t cltype) {
         do_test_unclean_destroy(cltype, 0/*tempq*/);
         do_test_unclean_destroy(cltype, 1/*mainq*/);
 
-        test_conf_init(&conf, NULL, 60);
+        test_conf_init(&conf, NULL, 180);
         test_conf_set(conf, "socket.timeout.ms", "10000");
         rk = test_create_handle(cltype, conf);
 
@@ -1183,6 +1939,27 @@ static void do_test_apis (rd_kafka_type_t cltype) {
         /* DescribeConfigs */
         do_test_DescribeConfigs(rk, mainq);
 
+        /* Delete records */
+        do_test_DeleteRecords("temp queue, op timeout 0", rk, NULL, 0);
+        do_test_DeleteRecords("main queue, op timeout 1500", rk, mainq, 1500);
+
+        /* Delete groups */
+        do_test_DeleteGroups("temp queue, op timeout 0", rk, NULL, 0);
+        do_test_DeleteGroups("main queue, op timeout 1500", rk, mainq, 1500);
+        do_test_DeleteGroups("main queue, op timeout 1500", rk, mainq, 1500);
+
+        if (test_broker_version >= TEST_BRKVER(2,4,0,0)) {
+                /* Delete committed offsets */
+                do_test_DeleteConsumerGroupOffsets(
+                        "temp queue, op timeout 0", rk, NULL, 0, rd_false);
+                do_test_DeleteConsumerGroupOffsets(
+                        "main queue, op timeout 1500", rk, mainq, 1500,
+                        rd_false);
+                do_test_DeleteConsumerGroupOffsets(
+                        "main queue, op timeout 1500", rk, mainq, 1500,
+                        rd_true/*with subscribing consumer*/);
+        }
+
         rd_kafka_queue_destroy(mainq);
 
         rd_kafka_destroy(rk);
@@ -1204,4 +1981,3 @@ int main_0081_admin (int argc, char **argv) {
 
         return 0;
 }
-
