@@ -64,7 +64,6 @@
  * Test output is controlled through "TEST_LEVEL=N" environemnt variable.
  * N < 2: TEST_SAY() is quiet.
  */
-extern int test_level;
 
 extern int test_seed;
 extern char test_mode[64];
@@ -145,6 +144,7 @@ struct test {
         test_state_t state;
         int     failcnt;     /**< Number of failures, useful with FAIL_LATER */
         char    failstr[512];/**< First test failure reason */
+        char    subtest[400];/**< Current subtest, if any */
 
 #if WITH_SOCKEM
         rd_list_t sockets;
@@ -444,6 +444,7 @@ rd_kafka_resp_err_t test_produce_sync (rd_kafka_t *rk, rd_kafka_topic_t *rkt,
 void test_produce_msgs_easy_v (const char *topic, uint64_t testid,
                                int32_t partition,
                                int msg_base, int cnt, size_t size, ...);
+void test_produce_msgs_easy_multi (uint64_t testid, ...);
 
 void test_rebalance_cb (rd_kafka_t *rk,
                         rd_kafka_resp_err_t err,
@@ -520,8 +521,13 @@ int test_consumer_poll (const char *what, rd_kafka_t *rk, uint64_t testid,
 
 void test_consumer_wait_assignment (rd_kafka_t *rk);
 void test_consumer_assign (const char *what, rd_kafka_t *rk,
-			   rd_kafka_topic_partition_list_t *parts);
+                           rd_kafka_topic_partition_list_t *parts);
+void test_consumer_incremental_assign (const char *what, rd_kafka_t *rk,
+                                       rd_kafka_topic_partition_list_t *parts);
 void test_consumer_unassign (const char *what, rd_kafka_t *rk);
+void test_consumer_incremental_unassign (const char *what, rd_kafka_t *rk,
+                                         rd_kafka_topic_partition_list_t
+                                         *parts);
 void test_consumer_assign_partition (const char *what, rd_kafka_t *rk,
                                      const char *topic, int32_t partition,
                                      int64_t offset);
@@ -535,6 +541,8 @@ void test_flush (rd_kafka_t *rk, int timeout_ms);
 
 void test_conf_set (rd_kafka_conf_t *conf, const char *name, const char *val);
 char *test_conf_get (const rd_kafka_conf_t *conf, const char *name);
+char *test_topic_conf_get (const rd_kafka_topic_conf_t *tconf,
+                           const char *name);
 int test_conf_match (rd_kafka_conf_t *conf, const char *name, const char *val);
 void test_topic_conf_set (rd_kafka_topic_conf_t *tconf,
                           const char *name, const char *val);
@@ -544,6 +552,8 @@ void test_any_conf_set (rd_kafka_conf_t *conf,
 
 void test_print_partition_list (const rd_kafka_topic_partition_list_t
 				*partitions);
+int test_partition_list_cmp (rd_kafka_topic_partition_list_t *al,
+                             rd_kafka_topic_partition_list_t *bl);
 
 void test_kafka_topics (const char *fmt, ...);
 void test_create_topic (rd_kafka_t *use_rk,
@@ -630,6 +640,26 @@ test_AlterConfigs_simple (rd_kafka_t *rk,
                           const char *resname,
                           const char **configs, size_t config_cnt);
 
+rd_kafka_resp_err_t
+test_DeleteGroups_simple (rd_kafka_t *rk,
+                          rd_kafka_queue_t *useq,
+                          char **groups, size_t group_cnt,
+                          void *opaque);
+
+rd_kafka_resp_err_t
+test_DeleteRecords_simple (rd_kafka_t *rk,
+                           rd_kafka_queue_t *useq,
+                           const rd_kafka_topic_partition_list_t *offsets,
+                           void *opaque);
+
+rd_kafka_resp_err_t
+test_DeleteConsumerGroupOffsets_simple (
+        rd_kafka_t *rk,
+        rd_kafka_queue_t *useq,
+        const char *group_id,
+        const rd_kafka_topic_partition_list_t *offsets,
+        void *opaque);
+
 rd_kafka_resp_err_t test_delete_all_test_topics (int timeout_ms);
 
 
@@ -638,6 +668,11 @@ rd_kafka_mock_cluster_t *test_mock_cluster_new (int broker_cnt,
                                                 const char **bootstraps);
 
 
+
+int test_error_is_not_fatal_cb (rd_kafka_t *rk, rd_kafka_resp_err_t err,
+                                const char *reason);
+
+
 /**
  * @brief Calls rdkafka function (with arguments)
  *        and checks its return value (must be rd_kafka_resp_err_t) for
